from typing import Callable

from PyQt5.QtCore import QMetaObject, pyqtBoundSignal
from PyQt5.QtGui import QBrush, QPalette
from PyQt5.QtWidgets import QAbstractItemView, QAction, QWidget


def connect_current_changed_signal(
    view: QAbstractItemView, slot: Callable[..., None] | pyqtBoundSignal
) -> None:
    selection_model = view.selectionModel()

    if selection_model is None:
        raise ValueError('selection_model is None!')

    selection_model.currentChanged.connect(slot)


def connect_triggered_signal(
    action: QAction | None, slot: Callable[..., None] | pyqtBoundSignal
) -> QMetaObject.Connection:
    if action is None:
        raise ValueError('action is None!')

    return action.triggered.connect(slot)


def create_brush_for_editable_cell(widget: QWidget) -> QBrush:
    palette = widget.palette()
    alternate_base_color = palette.color(QPalette.AlternateBase)
    return QBrush(alternate_base_color)



import logging

from ptychodus.api.observer import Observable, Observer

from ...model.analysis import XMCDAnalyzer
from ...model.visualization import VisualizationEngine
from ...view.object import XMCDDialog
from ...view.widgets import ExceptionDialog
from ..data import FileDialogFactory
from ..visualization import (
    VisualizationParametersController,
    VisualizationWidgetController,
)
from .tree_model import ObjectTreeModel

logger = logging.getLogger(__name__)


class XMCDViewController(Observer):
    def __init__(
        self,
        analyzer: XMCDAnalyzer,
        engine: VisualizationEngine,
        file_dialog_factory: FileDialogFactory,
        tree_model: ObjectTreeModel,
    ) -> None:
        super().__init__()
        self._analyzer = analyzer
        self._engine = engine
        self._file_dialog_factory = file_dialog_factory
        self._dialog = XMCDDialog()
        self._dialog.setWindowTitle('X-ray Magnetic Circular Dichroism (XMCD)')
        self._dialog.parameters_view.lcirc_combo_box.setModel(tree_model)
        self._dialog.parameters_view.lcirc_combo_box.currentIndexChanged.connect(
            analyzer.set_lcirc_product
        )
        self._dialog.parameters_view.rcirc_combo_box.setModel(tree_model)
        self._dialog.parameters_view.rcirc_combo_box.currentIndexChanged.connect(
            analyzer.set_rcirc_product
        )
        self._dialog.parameters_view.save_button.clicked.connect(self._save_data)

        self._difference_visualization_widget_controller = VisualizationWidgetController(
            engine,
            self._dialog.difference_widget,
            self._dialog.status_bar,
            file_dialog_factory,
        )
        self._sum_visualization_widget_controller = VisualizationWidgetController(
            engine, self._dialog.sum_widget, self._dialog.status_bar, file_dialog_factory
        )
        self._ratio_visualization_widget_controller = VisualizationWidgetController(
            engine, self._dialog.ratio_widget, self._dialog.status_bar, file_dialog_factory
        )
        self._visualization_parameters_controller = VisualizationParametersController(
            engine, self._dialog.parameters_view.visualization_parameters_view
        )

        analyzer.add_observer(self)

    def analyze(self, lcirc_product_index: int, rcirc_product_index: int) -> None:
        self._analyzer.set_lcirc_product(lcirc_product_index)
        self._analyzer.set_rcirc_product(rcirc_product_index)
        self._analyzer.analyze()
        self._dialog.open()

    def _save_data(self) -> None:
        title = 'Save XMCD Data'
        file_path, _ = self._file_dialog_factory.get_save_file_path(
            self._dialog,
            title,
            name_filters=self._analyzer.get_save_file_filters(),
            selected_name_filter=self._analyzer.get_save_file_filter(),
        )

        if file_path:
            try:
                self._analyzer.save_data(file_path)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception(title, err)

    def _sync_model_to_view(self) -> None:
        lcirc_product_index = self._analyzer.get_lcirc_product()
        self._dialog.parameters_view.lcirc_combo_box.setCurrentIndex(lcirc_product_index)

        rcirc_product_index = self._analyzer.get_rcirc_product()
        self._dialog.parameters_view.rcirc_combo_box.setCurrentIndex(rcirc_product_index)

        try:
            data = self._analyzer.get_result()
        except ValueError:
            self._difference_visualization_widget_controller.clear_array()
            self._sum_visualization_widget_controller.clear_array()
            self._ratio_visualization_widget_controller.clear_array()
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Update Views', err)
        else:
            self._difference_visualization_widget_controller.set_array(
                data.polar_difference, data.pixel_geometry
            )
            self._sum_visualization_widget_controller.set_array(data.polar_sum, data.pixel_geometry)
            self._ratio_visualization_widget_controller.set_array(
                data.polar_ratio, data.pixel_geometry
            )

    def _update(self, observable: Observable) -> None:
        if observable is self._analyzer:
            self._sync_model_to_view()



from pathlib import Path

import numpy

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe import ProbeSequence, ProbeFileReader, ProbeFileWriter


class NPYProbeFileReader(ProbeFileReader):
    def read(self, file_path: Path) -> ProbeSequence:
        array = numpy.load(file_path)
        return ProbeSequence(array=array, opr_weights=None, pixel_geometry=None)


class NPYProbeFileWriter(ProbeFileWriter):
    def write(self, file_path: Path, probes: ProbeSequence) -> None:
        array = probes.get_probe_no_opr().get_array()
        numpy.save(file_path, array)


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_file_readers.register_plugin(
        NPYProbeFileReader(),
        simple_name='NPY',
        display_name='NumPy Binary Files (*.npy)',
    )
    registry.probe_file_writers.register_plugin(
        NPYProbeFileWriter(),
        simple_name='NPY',
        display_name='NumPy Binary Files (*.npy)',
    )



from __future__ import annotations
from collections.abc import Sequence
from dataclasses import dataclass
import logging

import numpy
import scipy.fft

from ptychodus.api.common import ComplexArrayType, IntegerArrayType
from ptychodus.api.visualization import Plot2D, PlotAxis, PlotSeries

from ..product import ObjectRepository

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class FourierRingCorrelation:
    spatial_frequency_per_m: Sequence[float]
    correlation: Sequence[float]

    def get_resolution_m(self, threshold: float) -> float:
        # TODO threshold from bits
        for freq_rm, frc in zip(self.spatial_frequency_per_m, self.correlation):
            if frc < threshold:
                return 1.0 / freq_rm

        return numpy.nan

    def get_plot(self) -> Plot2D:
        freq_series = PlotSeries('freq', [1.0e-9 * freq for freq in self.spatial_frequency_per_m])
        frc_series = PlotSeries('frc', self.correlation)

        return Plot2D(
            axis_x=PlotAxis('Spatial Frequency [1/nm]', [freq_series]),
            axis_y=PlotAxis('Fourier Ring Correlation', [frc_series]),
        )


class FourierRingCorrelator:
    def __init__(self, repository: ObjectRepository) -> None:
        self._repository = repository

    @staticmethod
    def _integrate_rings(rings: IntegerArrayType, array: ComplexArrayType) -> ComplexArrayType:
        total = numpy.zeros(numpy.max(rings) + 1, dtype=complex)

        for index, value in zip(rings.flat, array.flat):
            total[index] += value

        return total

    def correlate(self, product_index_1: int, product_index_2: int) -> FourierRingCorrelation:
        """
        See: Joan Vila-Comamala, Ana Diaz, Manuel Guizar-Sicairos, Alexandre Mantion,
        Cameron M. Kewish, Andreas Menzel, Oliver Bunk, and Christian David,
        "Characterization of high-resolution diffractive X-ray optics by ptychographic
        coherent diffractive imaging," Opt. Express 19, 21333-21344 (2011)
        """

        object1 = self._repository[product_index_1].get_object()
        object2 = self._repository[product_index_2].get_object()

        # TODO support multilayer objects
        array1 = object1.get_layer(0)
        array2 = object2.get_layer(0)

        if numpy.ndim(array1) != 2 or numpy.ndim(array2) != 2:
            raise ValueError('Arrays must be 2D!')

        if numpy.shape(array1) != numpy.shape(array2):
            raise ValueError('Arrays must have same shape!')

        # TODO verify compatible pixel geometry
        pixel_geometry = object2.get_pixel_geometry()

        # TODO subpixel image registration: skimage.registration.phase_cross_correlation
        # TODO remove phase offset and ramp
        # TODO apply soft-edged mask
        # TODO stats: SSNR, area under FRC curve, average SNR, etc.

        x_rm = scipy.fft.fftfreq(array1.shape[-1], d=pixel_geometry.width_m)
        y_rm = scipy.fft.fftfreq(array1.shape[-2], d=pixel_geometry.height_m)
        radial_bin_size_per_m = max(x_rm[1], y_rm[1])

        xx_rm, yy_rm = numpy.meshgrid(x_rm, y_rm)
        rr_rm = numpy.hypot(xx_rm, yy_rm)

        rings = numpy.divide(rr_rm, radial_bin_size_per_m).astype(int)
        spatial_frequency_per_m = numpy.arange(numpy.max(rings) + 1) * radial_bin_size_per_m

        sf1 = scipy.fft.fft2(array1)
        sf2 = scipy.fft.fft2(array2)

        c11 = self._integrate_rings(rings, numpy.multiply(sf1, numpy.conj(sf1)))
        c12 = self._integrate_rings(rings, numpy.multiply(sf1, numpy.conj(sf2)))
        c22 = self._integrate_rings(rings, numpy.multiply(sf2, numpy.conj(sf2)))

        correlation = numpy.absolute(c12) / numpy.sqrt(numpy.absolute(numpy.multiply(c11, c22)))

        # TODO replace NaNs with interpolated values

        rnyquist = numpy.min(array1.shape) // 2 + 1
        return FourierRingCorrelation(
            spatial_frequency_per_m[:rnyquist].tolist(), correlation[:rnyquist].tolist()
        )



import skimage.restoration

from ptychodus.api.fluorescence import DeconvolutionStrategy, ElementMap
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.product import Product


class IdentityDeconvolution(DeconvolutionStrategy):
    def __call__(self, emap: ElementMap, product: Product) -> ElementMap:
        return emap


class RichardsonLucyDeconvolution(DeconvolutionStrategy):
    def __call__(self, emap: ElementMap, product: Product) -> ElementMap:
        probe_intensity = product.probes.get_probe_no_opr().get_intensity()
        cps = skimage.restoration.richardson_lucy(emap.counts_per_second, probe_intensity)
        return ElementMap(emap.name, cps)


class WienerDeconvolution(DeconvolutionStrategy):
    def __call__(self, emap: ElementMap, product: Product) -> ElementMap:
        probe_intensity = product.probes.get_probe_no_opr().get_intensity()
        balance = 0.05  # TODO
        cps = skimage.restoration.wiener(emap.counts_per_second, probe_intensity, balance)
        return ElementMap(emap.name, cps)


class UnsupervisedWienerDeconvolution(DeconvolutionStrategy):
    def __call__(self, emap: ElementMap, product: Product) -> ElementMap:
        probe_intensity = product.probes.get_probe_no_opr().get_intensity()
        cps, _ = skimage.restoration.unsupervised_wiener(emap.counts_per_second, probe_intensity)
        return ElementMap(emap.name, cps)


def register_plugins(registry: PluginRegistry) -> None:
    # NOTE See https://scikit-image.org/docs/stable/api/skimage.restoration.html
    # TODO Implement method from https://doi.org/10.1364/OE.20.018287
    registry.deconvolution_strategies.register_plugin(
        IdentityDeconvolution(),
        display_name='Identity',
    )
    registry.deconvolution_strategies.register_plugin(
        RichardsonLucyDeconvolution(),
        display_name='Richardson-Lucy',
    )
    registry.deconvolution_strategies.register_plugin(
        WienerDeconvolution(),
        display_name='Wiener',
    )
    registry.deconvolution_strategies.register_plugin(
        UnsupervisedWienerDeconvolution(),
        display_name='Unsupervised Wiener',
    )



from collections.abc import Iterator, Mapping, Sequence
from pathlib import Path
import logging
import threading
import time

from ptychodus.api.reconstructor import (
    NullReconstructor,
    PositionIndexFilter,
    ReconstructInput,
    Reconstructor,
    ReconstructorLibrary,
    TrainableReconstructor,
)

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import Parameter, StringParameter

from ..diffraction import DiffractionAPI
from ..product import ProductAPI
from ..task_manager import TaskManager
from .context import (
    ProcessingContext,
    ProcessingProgressMonitor,
    ReconstructBackgroundTask,
    TrainBackgroundTask,
)

logger = logging.getLogger(__name__)


class ProcessingAlgorithmParameter(Parameter[str], Observer):
    @staticmethod
    def generate_key(library: str, reconstructor: str) -> str:
        return f'{library.casefold()}_{reconstructor.casefold()}'

    @staticmethod
    def split_key(key: str) -> tuple[str, str]:
        library, reconstructor = key.split('_', maxsplit=1)
        return library, reconstructor

    def __init__(
        self, parameter: StringParameter, libraries: Sequence[ReconstructorLibrary]
    ) -> None:
        super().__init__()
        self._parameter = parameter
        self._libraries = libraries
        self._reconstructor_map: Mapping[str, Reconstructor] = {
            self.generate_key(library.name, reconstructor.name): reconstructor
            for library in libraries
            for reconstructor in library
        }
        self._current_value = 'none_none'
        self._current_reconstructor: Reconstructor = NullReconstructor('None')

        parameter.add_observer(self)
        self.set_value(parameter.get_value())

    def get_value(self) -> str:
        return self._current_value

    def set_value(self, value: str, *, notify: bool = True) -> None:
        sanitized_value = value.strip().casefold()

        if sanitized_value != self._current_value:
            try:
                reconstructor = self._reconstructor_map[sanitized_value]
            except KeyError:
                logger.debug(
                    f'Invalid plugin name "{value}". '
                    f'Registered plugins: {self._reconstructor_map.keys()}.'
                )
                self._parameter.set_value(self._current_value)
            else:
                logger.debug(f'{self._current_value} -> {sanitized_value}')
                self._current_value = sanitized_value
                self._current_reconstructor = reconstructor
                self._parameter.set_value(self._current_value)

                if notify:
                    self.notify_observers()

    def get_value_as_string(self) -> str:
        return self.get_value()

    def set_value_from_string(self, value: str) -> None:
        self.set_value(value)

    def available_reconstructors(self) -> Iterator[str]:
        for key in self._reconstructor_map.keys():
            yield key

    def get_current_reconstructor(self) -> Reconstructor:
        return self._current_reconstructor

    def copy(self) -> Parameter[str]:
        return ProcessingAlgorithmParameter(self._parameter, self._libraries)

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self.set_value(self._parameter.get_value())


class ProcessingAPI:
    def __init__(
        self,
        task_manager: TaskManager,
        diffraction_api: DiffractionAPI,
        product_api: ProductAPI,
        algorithm: ProcessingAlgorithmParameter,
        context: ProcessingContext,
    ) -> None:
        self._task_manager = task_manager
        self._diffraction_api = diffraction_api
        self._product_api = product_api
        self._algorithm_parameter = algorithm
        self._context = context

    def get_progress_monitor(self) -> ProcessingProgressMonitor:
        return self._context.get_progress_monitor()

    def get_reconstruct_input(
        self,
        product_index: int,
        *,
        index_filter: PositionIndexFilter = PositionIndexFilter.ALL,
    ) -> ReconstructInput:
        product = self._product_api.get_item(product_index).get_product()
        logger.info(f'Preparing input data for {product.metadata.name}...')
        tic = time.perf_counter()
        assembled_data = self._diffraction_api.get_assembled_data()
        reconstruct_input = assembled_data.prepare_reconstruct_input(
            product, index_filter=index_filter
        )
        toc = time.perf_counter()
        logger.info(f'Data preparation time {toc - tic:.4f} seconds.')
        return reconstruct_input

    def reconstruct(
        self,
        input_product_index: int,
        *,
        algorithm: str | None = None,
        index_filter: PositionIndexFilter = PositionIndexFilter.ALL,
        output_product_suffix: str = '',
        output_product_file: Path | None = None,
        block: bool = False,
    ) -> int:
        self.set_reconstructor_if_provided(algorithm)
        input_product_item = self._product_api.get_item(input_product_index)
        output_product_index = self._product_api.insert_product(input_product_item.get_product())
        output_product_item = self._product_api.get_item(output_product_index)
        output_product_name = (
            f'{input_product_item.get_name()}_{self._algorithm_parameter.get_value()}'
        )

        if output_product_suffix:
            output_product_name += f'_{output_product_suffix}'

        output_product_item.set_name(output_product_name)
        reconstruct_input = self.get_reconstruct_input(
            output_product_index, index_filter=index_filter
        )
        finished_event = threading.Event()
        background_task = ReconstructBackgroundTask(
            self._context,
            self._algorithm_parameter.get_current_reconstructor(),
            reconstruct_input,
            output_product_item,
            output_product_file,
            finished_event,
        )
        self._task_manager.put_background_task(background_task)

        if block:
            while not self._task_manager.is_stopping:
                if finished_event.wait(timeout=TaskManager.WAIT_TIME_S):
                    self._task_manager.run_foreground_tasks()
                    break

        return output_product_index

    def reconstruct_split(
        self, input_product_index: int, *, algorithm: str | None = None
    ) -> tuple[int, int]:
        output_product_index_odd = self.reconstruct(
            input_product_index,
            algorithm=algorithm,
            index_filter=PositionIndexFilter.ODD,
            output_product_suffix='odd',
        )
        output_product_index_even = self.reconstruct(
            input_product_index,
            algorithm=algorithm,
            index_filter=PositionIndexFilter.EVEN,
            output_product_suffix='even',
        )

        return output_product_index_odd, output_product_index_even

    def load_model_from_file(self, file_path: Path, algorithm: str | None = None) -> None:
        self.set_reconstructor_if_provided(algorithm)
        reconstructor = self._algorithm_parameter.get_current_reconstructor()

        if isinstance(reconstructor, TrainableReconstructor):
            logger.info('Opening model...')
            tic = time.perf_counter()
            reconstructor.load_model_from_file(file_path)
            toc = time.perf_counter()
            logger.info(f'Open time {toc - tic:.4f} seconds.')
        else:
            logger.warning('Algorithm is not trainable!')

    def export_training_data(
        self,
        file_path: Path,
        product_index: int,
        algorithm: str | None = None,
        index_filter: PositionIndexFilter = PositionIndexFilter.ALL,
    ) -> None:
        self.set_reconstructor_if_provided(algorithm)
        trainer = self._algorithm_parameter.get_current_reconstructor()

        if isinstance(trainer, TrainableReconstructor):
            reconstruct_input = self.get_reconstruct_input(product_index, index_filter=index_filter)

            logger.info('Exporting...')
            tic = time.perf_counter()
            trainer.export_training_data(file_path, reconstruct_input)
            toc = time.perf_counter()
            logger.info(f'Export time {toc - tic:.4f} seconds.')
        else:
            logger.warning('Algorithm is not trainable!')

    def train(
        self,
        product_index: int,
        input_path: Path,
        output_path: Path,
        *,
        algorithm: str | None = None,
        block: bool = False,
    ) -> None:
        self.set_reconstructor_if_provided(algorithm)
        trainer = self._algorithm_parameter.get_current_reconstructor()

        if isinstance(trainer, TrainableReconstructor):
            reconstruct_input = self.get_reconstruct_input(product_index)
            finished_event = threading.Event()
            background_task = TrainBackgroundTask(
                self._context,
                trainer,
                reconstruct_input,
                input_path,
                output_path,
                finished_event,
            )
            self._task_manager.put_background_task(background_task)

            if block:
                while not self._task_manager.is_stopping:
                    if finished_event.wait(timeout=TaskManager.WAIT_TIME_S):
                        self._task_manager.run_foreground_tasks()
                        break
        else:
            logger.warning('Algorithm is not trainable!')

    def get_training_data_file_filter(self, *, algorithm: str | None = None) -> str:
        self.set_reconstructor_if_provided(algorithm)
        trainer = self._algorithm_parameter.get_current_reconstructor()

        if isinstance(trainer, TrainableReconstructor):
            return trainer.get_training_data_file_filter()
        else:
            logger.warning('Algorithm is not trainable!')

        return ''

    def get_model_file_filter(self, *, algorithm: str | None = None) -> str:
        self.set_reconstructor_if_provided(algorithm)
        trainer = self._algorithm_parameter.get_current_reconstructor()

        if isinstance(trainer, TrainableReconstructor):
            return trainer.get_model_file_filter()
        else:
            logger.warning('Algorithm is not trainable!')

        return ''

    def available_reconstructors(self) -> Iterator[str]:
        return self._algorithm_parameter.available_reconstructors()

    def available_reconstructors_parts(self) -> Iterator[tuple[str, str]]:
        for key in self._algorithm_parameter.available_reconstructors():
            yield self._algorithm_parameter.split_key(key)

    def set_reconstructor_if_provided(self, algorithm: str | None) -> None:
        if algorithm is not None:
            self._algorithm_parameter.set_value(algorithm)



from typing import Any

from PyQt5.QtCore import (
    QAbstractTableModel,
    QModelIndex,
    QObject,
    QSortFilterProxyModel,
    Qt,
)
from PyQt5.QtGui import QBrush
from PyQt5.QtWidgets import QWidget

from ptychodus.api.observer import Observable, Observer

from ...model.diffraction import DiffractionAPI
from ...model.product import ProductRepositoryItem
from ...view.product import ProductEditorDialog
from ..helpers import create_brush_for_editable_cell


class ProductPropertyTableModel(QAbstractTableModel):
    def __init__(
        self,
        product_item: ProductRepositoryItem,
        editable_item_brush: QBrush,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent)
        self._product_item = product_item
        self._editable_item_brush = editable_item_brush
        self._header = ['Property', 'Value']
        self._properties = [
            'Probe Wavelength [nm]',
            'Probe Wavenumber [1/nm]',
            'Probe Angular Wavenumber [rad/nm]',
            'Probe Photon Flux [ph/s]',
            'Probe Power [W]',
            'Object Plane Pixel Width [nm]',
            'Object Plane Pixel Height [nm]',
            'Exposure Time [s]',
            'Mass Attenuation [m\u00b2/kg]',
            'Tomography Angle [deg]',
            'Fresnel Number',
            'Detector Numerical Aperture',
            'Depth of Field [nm]',
        ]

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        value = super().flags(index)

        if index.isValid() and index.row() in (7, 8, 9):
            value |= Qt.ItemFlag.ItemIsEditable

        return value

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._header[section]

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole:
                match index.column():
                    case 0:
                        return self._properties[index.row()]
                    case 1:
                        metadata_item = self._product_item.get_metadata_item()
                        geometry = self._product_item.get_geometry()

                        match index.row():
                            case 0:
                                return f'{geometry.probe_wavelength_m * 1e9:.4g}'
                            case 1:
                                return f'{geometry.probe_wavelengths_per_m * 1e-9:.4g}'
                            case 2:
                                return f'{geometry.probe_radians_per_m * 1e-9:.4g}'
                            case 3:
                                return f'{geometry.probe_photons_per_s:.4g}'
                            case 4:
                                return f'{geometry.probe_power_W:.4g}'
                            case 5:
                                return f'{geometry.object_plane_pixel_width_m * 1e9:.4g}'
                            case 6:
                                return f'{geometry.object_plane_pixel_height_m * 1e9:.4g}'
                            case 7:
                                return f'{metadata_item.exposure_time_s.get_value():.4g}'
                            case 8:
                                return f'{metadata_item.mass_attenuation_m2_kg.get_value():.4g}'
                            case 9:
                                return f'{metadata_item.tomography_angle_deg.get_value():.4g}'
                            case 10:
                                try:
                                    return f'{geometry.fresnel_number:.4g}'
                                except ZeroDivisionError:
                                    return 'inf'
                            case 11:
                                try:
                                    return f'{geometry.detector_numerical_aperture:.4g}'
                                except ZeroDivisionError:
                                    return 'inf'
                            case 12:
                                try:
                                    return f'{geometry.depth_of_field_m * 1e9:.4g}'
                                except ZeroDivisionError:
                                    return 'inf'
            elif role == Qt.ItemDataRole.BackgroundRole:
                if index.flags() & Qt.ItemFlag.ItemIsEditable:
                    return self._editable_item_brush

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:  # noqa: N802
        if index.isValid() and role == Qt.ItemDataRole.EditRole:
            metadata_item = self._product_item.get_metadata_item()

            match index.row():
                case 7:
                    try:
                        exposure_time_s = float(value)
                    except ValueError:
                        return False

                    metadata_item.exposure_time_s.set_value(exposure_time_s)
                    return True
                case 8:
                    try:
                        mass_attenuation_m2_kg = float(value)
                    except ValueError:
                        return False

                    metadata_item.mass_attenuation_m2_kg.set_value(mass_attenuation_m2_kg)
                    return True
                case 9:
                    try:
                        tomography_angle_deg = float(value)
                    except ValueError:
                        return False

                    metadata_item.tomography_angle_deg.set_value(tomography_angle_deg)
                    return True

        return False

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._properties)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._header)


class ProductEditorViewController(Observer):
    def __init__(
        self,
        diffraction_api: DiffractionAPI,
        product: ProductRepositoryItem,
        table_model: ProductPropertyTableModel,
        dialog: ProductEditorDialog,
    ) -> None:
        super().__init__()
        self._diffraction_api = diffraction_api
        self._product = product
        self._table_model = table_model
        self._dialog = dialog

    @classmethod
    def edit_product(
        cls, diffraction_api: DiffractionAPI, product: ProductRepositoryItem, parent: QWidget
    ) -> None:
        dialog = ProductEditorDialog(parent)
        dialog.setWindowTitle(f'Edit Product: {product.get_name()}')

        editable_item_brush = create_brush_for_editable_cell(dialog.table_view)
        table_model = ProductPropertyTableModel(product, editable_item_brush)

        table_proxy_model = QSortFilterProxyModel()
        table_proxy_model.setSourceModel(table_model)

        dialog.table_view.setModel(table_proxy_model)
        dialog.table_view.setSortingEnabled(True)
        vertical_header = dialog.table_view.verticalHeader()

        if vertical_header is not None:
            vertical_header.hide()

        dialog.table_view.resizeColumnsToContents()
        dialog.table_view.resizeRowsToContents()

        view_controller = cls(diffraction_api, product, table_model, dialog)
        product.add_observer(view_controller)
        dialog.text_edit.textChanged.connect(view_controller._sync_view_to_model)

        view_controller._sync_model_to_view()

        dialog.actions_view.estimate_probe_photon_count_button.clicked.connect(
            view_controller._estimate_probe_photon_count
        )
        dialog.finished.connect(view_controller._finish)
        dialog.open()
        dialog.adjustSize()

    def _sync_view_to_model(self) -> None:
        metadata = self._product.get_metadata_item()
        metadata.comments.set_value(self._dialog.text_edit.toPlainText())

    def _sync_model_to_view(self) -> None:
        self._table_model.beginResetModel()
        self._table_model.endResetModel()

        metadata = self._product.get_metadata_item()
        self._dialog.text_edit.setPlainText(metadata.comments.get_value())

    def _estimate_probe_photon_count(self) -> None:
        metadata = self._product.get_metadata_item()
        assembled_data = self._diffraction_api.get_assembled_data()
        metadata.probe_photon_count.set_value(assembled_data.get_pattern_counts().max())

        self._table_model.beginResetModel()
        self._table_model.endResetModel()

    def _finish(self, result: int) -> None:
        self._product.remove_observer(self)

    def _update(self, observable: Observable) -> None:
        if observable is self._product:
            self._sync_model_to_view()



from __future__ import annotations

import numpy

from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition

from .builder import ProbePositionsBuilder
from .settings import ProbePositionsSettings


class LissajousProbePositionsBuilder(ProbePositionsBuilder):
    def __init__(self, settings: ProbePositionsSettings) -> None:
        super().__init__(settings, 'lissajous')
        self._settings = settings

        self.num_points = settings.num_points_x.copy()
        self.num_points.set_value(
            settings.num_points_x.get_value() * settings.num_points_y.get_value()
        )
        self._add_parameter('num_points', self.num_points)

        self._num_points = settings.num_points_y.copy()
        self._num_points.set_value(1)
        self._add_parameter('_num_points', self._num_points)

        self.amplitude_x_m = settings.amplitude_x_m.copy()
        self._add_parameter('amplitude_x_m', self.amplitude_x_m)

        self.amplitude_y_m = settings.amplitude_y_m.copy()
        self._add_parameter('amplitude_y_m', self.amplitude_y_m)

        self.angular_step_x_turns = settings.angular_step_x_turns.copy()
        self._add_parameter('angular_step_x_tr', self.angular_step_x_turns)

        self.angular_step_y_turns = settings.angular_step_y_turns.copy()
        self._add_parameter('angular_step_y_tr', self.angular_step_y_turns)

        self.angular_shift_turns = settings.angular_shift_turns.copy()
        self._add_parameter('angular_shift_tr', self.angular_shift_turns)

    def copy(self) -> LissajousProbePositionsBuilder:
        builder = LissajousProbePositionsBuilder(self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        for index in range(self.num_points.get_value()):
            two_pi = 2 * numpy.pi
            theta_x = (
                two_pi * self.angular_step_x_turns.get_value() * index
                + self.angular_shift_turns.get_value()
            )
            theta_y = two_pi * self.angular_step_y_turns.get_value() * index

            point = ProbePosition(
                index=index,
                coordinate_x_m=self.amplitude_x_m.get_value() * numpy.sin(theta_x),
                coordinate_y_m=self.amplitude_y_m.get_value() * numpy.sin(theta_y),
            )
            point_list.append(point)

        return ProbePositionSequence(point_list)



from __future__ import annotations
from decimal import Decimal
import logging

import numpy

from PyQt5.QtCore import QStringListModel
from PyQt5.QtWidgets import QButtonGroup, QDialog, QStatusBar

from ptychodus.api.common import NumberArrayType
from ptychodus.api.geometry import Interval, PixelGeometry
from ptychodus.api.observer import Observable, Observer

from ..model.visualization import VisualizationEngine
from ..view.image import (
    ImageDataRangeGroupBox,
    ImageDisplayRangeDialog,
    ImageRendererGroupBox,
    ImageToolsGroupBox,
    ImageView,
    ImageWidget,
)
from ..view.visualization import ImageItem, ImageMouseTool
from .data import FileDialogFactory
from .visualization import VisualizationController

logger = logging.getLogger(__name__)


class ImageToolsController:
    MIME_TYPES = ['image/bmp', 'image/jpeg', 'image/png', 'image/x-portable-pixmap']

    def __init__(
        self,
        view: ImageToolsGroupBox,
        visualization_controller: VisualizationController,
    ) -> None:
        self._view = view
        self._visualization_controller = visualization_controller

        view.home_button.clicked.connect(visualization_controller.zoom_to_fit)
        view.save_button.clicked.connect(visualization_controller.save_image)
        view.move_button.setCheckable(True)
        view.move_button.setChecked(True)
        view.fourier_button.setCheckable(True)
        view.ruler_button.setCheckable(True)
        view.rectangle_button.setCheckable(True)
        view.line_cut_button.setCheckable(True)

        self._mouse_tool_button_group = QButtonGroup()
        self._mouse_tool_button_group.addButton(view.move_button, ImageMouseTool.MOVE_TOOL.value)
        self._mouse_tool_button_group.addButton(
            view.fourier_button, ImageMouseTool.FOURIER_TOOL.value
        )
        self._mouse_tool_button_group.addButton(view.ruler_button, ImageMouseTool.RULER_TOOL.value)
        self._mouse_tool_button_group.addButton(
            view.rectangle_button, ImageMouseTool.RECTANGLE_TOOL.value
        )
        self._mouse_tool_button_group.addButton(
            view.line_cut_button, ImageMouseTool.LINE_CUT_TOOL.value
        )
        self._mouse_tool_button_group.idToggled.connect(self._set_mouse_tool)

    def _set_mouse_tool(self, tool_id: int, checked: bool) -> None:
        if checked:
            mouse_tool = ImageMouseTool(tool_id)
            self._visualization_controller.set_mouse_tool(mouse_tool)


class ImageRendererController(Observer):
    def __init__(self, engine: VisualizationEngine, view: ImageRendererGroupBox) -> None:
        super().__init__()
        self._engine = engine
        self._view = view

        self._renderer_model = QStringListModel()
        view.renderer_combo_box.setModel(self._renderer_model)

        self._transformation_model = QStringListModel()
        view.transformation_combo_box.setModel(self._transformation_model)

        self._variant_model = QStringListModel()
        view.variant_combo_box.setModel(self._variant_model)

        self._sync_model_to_view()
        engine.add_observer(self)

        view.renderer_combo_box.textActivated.connect(engine.set_renderer)
        view.transformation_combo_box.textActivated.connect(engine.set_transformation)
        view.variant_combo_box.textActivated.connect(engine.set_variant)

    def _sync_model_to_view(self) -> None:
        self._view.renderer_combo_box.blockSignals(True)
        self._renderer_model.setStringList([name for name in self._engine.renderers()])
        self._view.renderer_combo_box.setCurrentText(self._engine.get_renderer())
        self._view.renderer_combo_box.blockSignals(False)

        self._view.transformation_combo_box.blockSignals(True)
        self._transformation_model.setStringList([name for name in self._engine.transformations()])
        self._view.transformation_combo_box.setCurrentText(self._engine.get_transformation())
        self._view.transformation_combo_box.blockSignals(False)

        self._view.variant_combo_box.blockSignals(True)
        self._variant_model.setStringList([name for name in self._engine.variants()])
        self._view.variant_combo_box.setCurrentText(self._engine.get_variant())
        self._view.variant_combo_box.blockSignals(False)

    def _update(self, observable: Observable) -> None:
        if observable is self._engine:
            self._sync_model_to_view()


class ImageDataRangeController(Observer):
    def __init__(
        self,
        engine: VisualizationEngine,
        view: ImageDataRangeGroupBox,
        image_widget: ImageWidget,
        visualization_controller: VisualizationController,
    ) -> None:
        self._engine = engine
        self._view = view
        self._image_widget = image_widget
        self._display_range_dialog = ImageDisplayRangeDialog(view)
        self._visualization_controller = visualization_controller
        self._display_range_is_locked = True

        self._sync_model_to_view()
        engine.add_observer(self)

        view.min_display_value_slider.value_changed.connect(
            lambda value: engine.set_min_display_value(float(value))
        )
        view.max_display_value_slider.value_changed.connect(
            lambda value: engine.set_max_display_value(float(value))
        )
        view.auto_button.clicked.connect(self._auto_display_range)
        view.edit_button.clicked.connect(self._display_range_dialog.open)
        self._display_range_dialog.finished.connect(self._finish_editing_display_range)

        view.color_legend_button.setCheckable(True)
        image_widget.set_color_legend_visible(view.color_legend_button.isChecked())
        view.color_legend_button.toggled.connect(image_widget.set_color_legend_visible)

    def _auto_display_range(self) -> None:
        self._display_range_is_locked = False
        self._visualization_controller.rerender_image(autoscale_color_axis=True)
        self._display_range_is_locked = True

    def _finish_editing_display_range(self, result: int) -> None:
        if result == QDialog.DialogCode.Accepted:
            lower = float(self._display_range_dialog.min_value_line_edit.get_value())
            upper = float(self._display_range_dialog.max_value_line_edit.get_value())

            self._display_range_is_locked = False
            self._engine.set_display_value_range(lower, upper)
            self._display_range_is_locked = True

    def _sync_color_legend_to_view(self) -> None:
        values = numpy.linspace(
            self._engine.get_min_display_value(), self._engine.get_max_display_value(), 1000
        )
        self._image_widget.set_color_legend_colors(
            values,
            self._engine.colorize(values),
            self._engine.is_renderer_cyclic(),
        )

    def _sync_model_to_view(self) -> None:
        min_value = Decimal(repr(self._engine.get_min_display_value()))
        max_value = Decimal(repr(self._engine.get_max_display_value()))

        self._display_range_dialog.min_value_line_edit.set_value(min_value)
        self._display_range_dialog.max_value_line_edit.set_value(max_value)

        if self._display_range_is_locked:
            self._view.min_display_value_slider.set_value(min_value)
            self._view.max_display_value_slider.set_value(max_value)
        else:
            display_range_limits = Interval[Decimal](min_value, max_value)
            self._view.min_display_value_slider.set_value_and_range(min_value, display_range_limits)
            self._view.max_display_value_slider.set_value_and_range(max_value, display_range_limits)

        self._sync_color_legend_to_view()

    def _update(self, observable: Observable) -> None:
        if observable is self._engine:
            self._sync_model_to_view()


class ImageController:
    def __init__(
        self,
        engine: VisualizationEngine,
        view: ImageView,
        status_bar: QStatusBar,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        self._visualization_controller = VisualizationController(
            engine, view.image_widget, status_bar, file_dialog_factory
        )
        self._tools_controller = ImageToolsController(
            view.image_ribbon.image_tools_group_box, self._visualization_controller
        )
        self._renderer_controller = ImageRendererController(
            engine, view.image_ribbon.colormap_group_box
        )
        self._data_range_controller = ImageDataRangeController(
            engine,
            view.image_ribbon.data_range_group_box,
            view.image_widget,
            self._visualization_controller,
        )

    def get_item(self) -> ImageItem:
        return self._visualization_controller.get_item()

    def set_array(self, array: NumberArrayType, pixel_geometry: PixelGeometry) -> None:
        self._visualization_controller.set_array(array, pixel_geometry)

    def clear_array(self) -> None:
        self._visualization_controller.clear_array()



from __future__ import annotations
from pathlib import Path

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class AutomationSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Automation')
        self._group.add_observer(self)

        self.workflow = self._group.create_string_parameter('Workflow', 'Autoload_Product')
        self.data_directory = self._group.create_path_parameter(
            'DataDirectory', Path('/path/to/data')
        )
        self.use_watchdog_polling_observer = self._group.create_boolean_parameter(
            'UseWatchdogPollingObserver', False
        )
        self.watchdog_delay_s = self._group.create_integer_parameter(
            'WatchdogDelayInSeconds', 15, minimum=0, maximum=3600
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from typing import Final
import traceback

from PyQt5.QtCore import QEvent
from PyQt5.QtWidgets import QMessageBox, QSizePolicy, QTextEdit


class ExceptionDialog(QMessageBox):
    MIN_SIZE: Final[int] = 0
    MAX_SIZE: Final[int] = 16777215

    @classmethod
    def show_exception(cls, actor: str, exception: Exception) -> None:
        dialog = cls()
        dialog.setSizeGripEnabled(True)
        dialog.setWindowTitle('Exception Dialog')
        dialog.setIcon(QMessageBox.Icon.Critical)
        dialog.setText(f'{actor} raised a {exception.__class__.__name__}!')
        dialog.setInformativeText(str(exception))
        dialog.setDetailedText(traceback.format_exc())
        _ = dialog.exec()

    def event(self, e: QEvent) -> bool:
        result = super().event(e)

        if e.type() in (QEvent.Type.LayoutRequest, QEvent.Type.Resize):
            self.setMinimumHeight(self.MIN_SIZE)
            self.setMaximumHeight(self.MAX_SIZE)
            self.setMinimumWidth(self.MIN_SIZE)
            self.setMaximumWidth(self.MAX_SIZE)
            self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

            text_edit = self.findChild(QTextEdit)

            if text_edit is not None:
                # make the detailed text expandable
                text_edit.setMinimumHeight(self.MIN_SIZE)
                text_edit.setMaximumHeight(self.MAX_SIZE)
                text_edit.setMinimumWidth(self.MIN_SIZE)
                text_edit.setMaximumWidth(self.MAX_SIZE)
                text_edit.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

        return result



"""I/O utilities for assembled diffraction data and data products."""

from __future__ import annotations
from enum import StrEnum
from pathlib import Path
import logging

import h5py

from .geometry import PixelGeometry
from .object import Object, ObjectCenter
from .probe import ProbeSequence
from .probe_positions import ProbePositionSequence, ProbePosition
from .product import Product, ProductMetadata
from .reconstructor import AssembledDiffractionData, LossValue

__all__ = [
    'StandardFileLayout',
    'load_diffraction_data',
    'load_product',
    'save_diffraction_data',
    'save_product',
]

logger = logging.getLogger(__name__)


class StandardFileLayout(StrEnum):
    DIFFRACTION = 'diffraction.h5'
    FLUORESCENCE_IN = 'fluorescence-in.h5'
    FLUORESCENCE_OUT = 'fluorescence-out.h5'
    PRODUCT_IN = 'product-in.h5'
    PRODUCT_OUT = 'product-out.h5'
    SETTINGS = 'settings.ini'


class DiffractionFileKeys(StrEnum):
    PATTERNS = 'patterns'
    INDEXES = 'indexes'
    BAD_PIXELS = 'bad_pixels'


def load_diffraction_data(file: Path, *, mmap_file: Path | None = None) -> AssembledDiffractionData:
    if mmap_file is not None:
        raise NotImplementedError('Load to memory map not implemented yet!')

    with h5py.File(file, 'r') as h5_file:
        h5_indexes = h5_file[DiffractionFileKeys.INDEXES]

        if not isinstance(h5_indexes, h5py.Dataset):
            raise ValueError('Indexes are not a dataset!')

        h5_patterns = h5_file[DiffractionFileKeys.PATTERNS]

        if not isinstance(h5_patterns, h5py.Dataset):
            raise ValueError('Patterns are not a dataset!')

        h5_bad_pixels = h5_file[DiffractionFileKeys.BAD_PIXELS]

        if not isinstance(h5_bad_pixels, h5py.Dataset):
            raise ValueError('Bad pixels are not a dataset!')

        return AssembledDiffractionData(
            h5_indexes[()],
            h5_patterns[()],
            h5_bad_pixels[()],
        )


def save_diffraction_data(
    file: Path, data: AssembledDiffractionData, *, compression: str = 'lzf'
) -> None:
    with h5py.File(file, 'w') as h5_file:
        h5_file.create_dataset(
            DiffractionFileKeys.INDEXES, data=data._indexes, compression=compression
        )
        h5_file.create_dataset(
            DiffractionFileKeys.PATTERNS, data=data._patterns, compression=compression
        )
        h5_file.create_dataset(
            DiffractionFileKeys.BAD_PIXELS, data=data._bad_pixels, compression=compression
        )


class ProductFileKeys(StrEnum):
    NAME = 'name'
    COMMENTS = 'comments'
    DETECTOR_OBJECT_DISTANCE = 'detector_object_distance_m'
    PROBE_ENERGY = 'probe_energy_eV'
    PROBE_PHOTON_COUNT = 'probe_photon_count'
    EXPOSURE_TIME = 'exposure_time_s'
    MASS_ATTENUATION = 'mass_attenuation_m2_kg'
    TOMOGRAPHY_ANGLE = 'tomography_angle_deg'
    PROBE_ARRAY = 'probe'
    OPR_WEIGHTS = 'opr_weights'
    PROBE_PIXEL_HEIGHT = 'pixel_height_m'
    PROBE_PIXEL_WIDTH = 'pixel_width_m'
    PROBE_POSITION_INDEXES = 'probe_position_indexes'
    PROBE_POSITION_X = 'probe_position_x_m'
    PROBE_POSITION_Y = 'probe_position_y_m'
    OBJECT_ARRAY = 'object'
    OBJECT_CENTER_X = 'center_x_m'
    OBJECT_CENTER_Y = 'center_y_m'
    OBJECT_LAYER_SPACING = 'object_layer_spacing_m'
    OBJECT_PIXEL_HEIGHT = 'pixel_height_m'
    OBJECT_PIXEL_WIDTH = 'pixel_width_m'
    LOSS_EPOCHS = 'loss_epochs'
    LOSS_VALUES = 'loss_values'


def load_product(file: Path) -> Product:
    point_list: list[ProbePosition] = []

    with h5py.File(file, 'r') as h5_file:
        name = str(h5_file.attrs.get(ProductFileKeys.NAME, 'Unnamed'))
        comments = str(h5_file.attrs.get(ProductFileKeys.COMMENTS, ''))
        probe_photon_count = int(h5_file.attrs.get(ProductFileKeys.PROBE_PHOTON_COUNT, 0))
        exposure_time_s = float(h5_file.attrs.get(ProductFileKeys.EXPOSURE_TIME, 0.0))
        mass_attenuation_m2_kg = float(h5_file.attrs.get(ProductFileKeys.MASS_ATTENUATION, 0.0))
        tomography_angle_deg = float(h5_file.attrs.get(ProductFileKeys.TOMOGRAPHY_ANGLE, 0.0))

        metadata = ProductMetadata(
            name=name,
            comments=comments,
            detector_distance_m=float(h5_file.attrs[ProductFileKeys.DETECTOR_OBJECT_DISTANCE]),
            probe_energy_eV=float(h5_file.attrs[ProductFileKeys.PROBE_ENERGY]),
            probe_photon_count=probe_photon_count,
            exposure_time_s=exposure_time_s,
            mass_attenuation_m2_kg=mass_attenuation_m2_kg,
            tomography_angle_deg=tomography_angle_deg,
        )

        h5_object = h5_file[ProductFileKeys.OBJECT_ARRAY]

        if not isinstance(h5_object, h5py.Dataset):
            raise ValueError('Object array is not a dataset!')

        object_pixel_geometry = PixelGeometry(
            width_m=float(h5_object.attrs[ProductFileKeys.OBJECT_PIXEL_WIDTH]),
            height_m=float(h5_object.attrs[ProductFileKeys.OBJECT_PIXEL_HEIGHT]),
        )
        object_center = ObjectCenter(
            coordinate_x_m=float(h5_object.attrs[ProductFileKeys.OBJECT_CENTER_X]),
            coordinate_y_m=float(h5_object.attrs[ProductFileKeys.OBJECT_CENTER_Y]),
        )

        try:
            h5_object_layer_spacing = h5_file[ProductFileKeys.OBJECT_LAYER_SPACING]
        except KeyError:
            logger.debug('Object layer spacing not found.')
            layer_spacing_m = []
        else:
            if isinstance(h5_object_layer_spacing, h5py.Dataset):
                layer_spacing_m = h5_object_layer_spacing[()]
            else:
                raise ValueError('Object layer spacing is not a dataset!')

        object_ = Object(
            array=h5_object[()],
            pixel_geometry=object_pixel_geometry,
            center=object_center,
            layer_spacing_m=layer_spacing_m,
        )

        h5_probe = h5_file[ProductFileKeys.PROBE_ARRAY]

        if not isinstance(h5_probe, h5py.Dataset):
            raise ValueError('Probe array is not a dataset!')

        probe_pixel_geometry = PixelGeometry(
            width_m=float(
                h5_probe.attrs.get(ProductFileKeys.PROBE_PIXEL_WIDTH, object_pixel_geometry.width_m)
            ),
            height_m=float(
                h5_probe.attrs.get(
                    ProductFileKeys.PROBE_PIXEL_HEIGHT, object_pixel_geometry.height_m
                )
            ),
        )

        try:
            h5_opr_weights = h5_file[ProductFileKeys.OPR_WEIGHTS]
        except KeyError:
            logger.debug('OPR weights not found.')
            opr_weights = None
        else:
            if isinstance(h5_opr_weights, h5py.Dataset):
                opr_weights = h5_opr_weights[()]
            else:
                raise ValueError('OPR weights is not a dataset!')

        probe = ProbeSequence(
            array=h5_probe[()],
            opr_weights=opr_weights,
            pixel_geometry=probe_pixel_geometry,
        )

        h5_position_indexes = h5_file[ProductFileKeys.PROBE_POSITION_INDEXES]

        if not isinstance(h5_position_indexes, h5py.Dataset):
            raise ValueError('Probe position indexes is not a dataset!')

        h5_position_x = h5_file[ProductFileKeys.PROBE_POSITION_X]

        if not isinstance(h5_position_x, h5py.Dataset):
            raise ValueError('Probe position X is not a dataset!')

        h5_position_y = h5_file[ProductFileKeys.PROBE_POSITION_Y]

        if not isinstance(h5_position_y, h5py.Dataset):
            raise ValueError('Probe position Y is not a dataset!')

        for idx, x_m, y_m in zip(h5_position_indexes[()], h5_position_x[()], h5_position_y[()]):
            point = ProbePosition(idx, x_m, y_m)
            point_list.append(point)

        losses: list[LossValue] = []

        try:
            h5_loss_epochs = h5_file[ProductFileKeys.LOSS_EPOCHS]
            h5_loss_values = h5_file[ProductFileKeys.LOSS_VALUES]
        except KeyError:
            logger.debug('Losses not found!')
        else:
            if not isinstance(h5_loss_epochs, h5py.Dataset):
                raise ValueError('Loss epochs are not a dataset!')

            if not isinstance(h5_loss_values, h5py.Dataset):
                raise ValueError('Loss values are not a dataset!')

            for epoch, value in zip(h5_loss_epochs[()], h5_loss_values[()]):
                loss = LossValue(epoch, value)
                losses.append(loss)

    return Product(
        metadata=metadata,
        probe_positions=ProbePositionSequence(point_list),
        probes=probe,
        object_=object_,
        losses=losses,
    )


def save_product(file: Path, product: Product) -> None:
    scan_indexes: list[int] = []
    scan_x_m: list[float] = []
    scan_y_m: list[float] = []

    for point in product.probe_positions:
        scan_indexes.append(point.index)
        scan_x_m.append(point.coordinate_x_m)
        scan_y_m.append(point.coordinate_y_m)

    with h5py.File(file, 'w') as h5_file:
        metadata = product.metadata
        h5_file.attrs[ProductFileKeys.NAME] = metadata.name
        h5_file.attrs[ProductFileKeys.COMMENTS] = metadata.comments
        h5_file.attrs[ProductFileKeys.DETECTOR_OBJECT_DISTANCE] = metadata.detector_distance_m
        h5_file.attrs[ProductFileKeys.PROBE_ENERGY] = metadata.probe_energy_eV
        h5_file.attrs[ProductFileKeys.PROBE_PHOTON_COUNT] = metadata.probe_photon_count
        h5_file.attrs[ProductFileKeys.EXPOSURE_TIME] = metadata.exposure_time_s
        h5_file.attrs[ProductFileKeys.MASS_ATTENUATION] = metadata.mass_attenuation_m2_kg

        h5_file.create_dataset(ProductFileKeys.PROBE_POSITION_INDEXES, data=scan_indexes)
        h5_file.create_dataset(ProductFileKeys.PROBE_POSITION_X, data=scan_x_m)
        h5_file.create_dataset(ProductFileKeys.PROBE_POSITION_Y, data=scan_y_m)

        probe = product.probes
        h5_probe = h5_file.create_dataset(ProductFileKeys.PROBE_ARRAY, data=probe.get_array())

        try:
            opr_weights = probe.get_opr_weights()
        except ValueError:
            pass
        else:
            h5_file.create_dataset(ProductFileKeys.OPR_WEIGHTS, data=opr_weights)

        probe_pixel_geometry = probe.get_pixel_geometry()
        h5_probe.attrs[ProductFileKeys.PROBE_PIXEL_WIDTH] = probe_pixel_geometry.width_m
        h5_probe.attrs[ProductFileKeys.PROBE_PIXEL_HEIGHT] = probe_pixel_geometry.height_m

        object_ = product.object_
        object_geometry = object_.get_geometry()
        h5_object = h5_file.create_dataset(ProductFileKeys.OBJECT_ARRAY, data=object_.get_array())
        h5_object.attrs[ProductFileKeys.OBJECT_CENTER_X] = object_geometry.center_x_m
        h5_object.attrs[ProductFileKeys.OBJECT_CENTER_Y] = object_geometry.center_y_m
        h5_object.attrs[ProductFileKeys.OBJECT_PIXEL_WIDTH] = object_geometry.pixel_width_m
        h5_object.attrs[ProductFileKeys.OBJECT_PIXEL_HEIGHT] = object_geometry.pixel_height_m
        h5_file.create_dataset(ProductFileKeys.OBJECT_LAYER_SPACING, data=object_.layer_spacing_m)

        loss_epochs: list[int] = []
        loss_values: list[float] = []

        for loss in product.losses:
            loss_epochs.append(loss.epoch)
            loss_values.append(loss.value)

        h5_file.create_dataset(ProductFileKeys.LOSS_EPOCHS, data=loss_epochs)
        h5_file.create_dataset(ProductFileKeys.LOSS_VALUES, data=loss_values)



from .core import FluorescenceCore, FluorescenceEnhancer
from .two_step import TwoStepFluorescenceEnhancingAlgorithm
from .vspi import VSPIFluorescenceEnhancingAlgorithm

__all__ = [
    'FluorescenceCore',
    'FluorescenceEnhancer',
    'TwoStepFluorescenceEnhancingAlgorithm',
    'VSPIFluorescenceEnhancingAlgorithm',
]



from typing import Any, overload

from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import Qt, QAbstractItemModel, QModelIndex, QObject

from ptychodus.api.tree import SimpleTreeNode

from ...model.diffraction import AssembledDiffractionDataset, DiffractionDatasetObserver
from ...view.diffraction import DatasetFileLayoutDialog


class SimpleTreeModel(QAbstractItemModel):
    def __init__(self, root_node: SimpleTreeNode, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._root_node = root_node

    def set_root_node(self, root_node: SimpleTreeNode) -> None:
        self.beginResetModel()
        self._root_node = root_node
        self.endResetModel()

    @overload
    def parent(self, child: QModelIndex) -> QModelIndex: ...

    @overload
    def parent(self) -> QObject: ...

    def parent(self, child: QModelIndex | None = None) -> QModelIndex | QObject:
        if child is None:
            return super().parent()
        else:
            value = QModelIndex()

            if child.isValid():
                child_item = child.internalPointer()
                parent_item = child_item.parent_item

                if parent_item is self._root_node:
                    value = QModelIndex()
                else:
                    value = self.createIndex(parent_item.row(), 0, parent_item)

            return value

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._root_node.data(section)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        return super().flags(index)

    def index(self, row: int, column: int, parent: QModelIndex = QModelIndex()) -> QModelIndex:
        value = QModelIndex()

        if self.hasIndex(row, column, parent):
            parent_item = parent.internalPointer() if parent.isValid() else self._root_node
            child_item = parent_item.child_items[row]

            if child_item:
                value = self.createIndex(row, column, child_item)

        return value

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid() and role == Qt.ItemDataRole.DisplayRole:
            node = index.internalPointer()
            return node.data(index.column())

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        if parent.column() > 0:
            return 0

        node = self._root_node

        if parent.isValid():
            node = parent.internalPointer()

        return len(node.child_items)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        node = self._root_node

        if parent.isValid():
            node = parent.internalPointer()

        return len(node.item_data)


class DatasetLayoutViewController(DiffractionDatasetObserver):
    def __init__(self, dataset: AssembledDiffractionDataset, tree_model: SimpleTreeModel) -> None:
        super().__init__()
        self._dataset = dataset
        self._tree_model = tree_model

    @classmethod
    def show_dialog(cls, dataset: AssembledDiffractionDataset, parent: QWidget) -> None:
        tree_model = SimpleTreeModel(dataset.get_layout())
        controller = cls(dataset, tree_model)
        dataset.add_observer(controller)

        dialog = DatasetFileLayoutDialog(parent)
        dialog.setWindowTitle('Dataset File Layout')
        dialog.tree_view.setModel(tree_model)

        controller._sync_model_to_view()
        dialog.open()

    def _sync_model_to_view(self) -> None:
        self._tree_model.set_root_node(self._dataset.get_layout())

    def handle_array_inserted(self, index: int) -> None:
        pass

    def handle_array_changed(self, index: int) -> None:
        pass

    def handle_dataset_reloaded(self) -> None:
        self._sync_model_to_view()



from __future__ import annotations
from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from pathlib import Path
from typing import overload
import logging

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent, PixelGeometry
from ptychodus.api.diffraction import (
    BadPixels,
    CropCenter,
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    DiffractionArray,
    SimpleDiffractionDataset,
)
from ptychodus.api.tree import SimpleTreeNode

from ..h5_diffraction_file import H5DiffractionPatternArray, H5DiffractionFileTreeBuilder

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class DataGroup:
    array_list: list[DiffractionArray] = field(default_factory=list)

    @classmethod
    def read(cls, group: h5py.Group, num_patterns_per_array: int) -> DataGroup:
        array_list: list[DiffractionArray] = list()
        master_file_path = Path(group.file.filename)

        for name, h5_item in sorted(group.items()):
            h5_item = group.get(name, getlink=True)

            if isinstance(h5_item, h5py.ExternalLink):
                array = H5DiffractionPatternArray(
                    label=name,
                    indexes=numpy.arange(num_patterns_per_array)
                    + len(array_list) * num_patterns_per_array,
                    file_path=master_file_path.parent / h5_item.filename,
                    data_path=str(h5_item.path),
                )
                array_list.append(array)

        return cls(array_list)

    def __iter__(self) -> Iterator[DiffractionArray]:
        return iter(self.array_list)

    @overload
    def __getitem__(self, index: int) -> DiffractionArray: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[DiffractionArray]: ...

    def __getitem__(self, index: int | slice) -> DiffractionArray | Sequence[DiffractionArray]:
        return self.array_list[index]

    def __len__(self) -> int:
        return len(self.array_list)


@dataclass(frozen=True)
class DetectorSpecificGroup:
    nimages: int
    ntrigger: int
    photon_energy_eV: float  # noqa: N815
    x_pixels_in_detector: int
    y_pixels_in_detector: int

    @property
    def num_patterns_total(self) -> int:
        return max(self.nimages, self.ntrigger)

    @classmethod
    def read(cls, group: h5py.Group) -> DetectorSpecificGroup:
        nimages = group['nimages']
        ntrigger = group['ntrigger']
        photon_energy = group['photon_energy']
        assert photon_energy.attrs['units'] == b'eV'
        x_pixels_in_detector = group['x_pixels_in_detector']
        y_pixels_in_detector = group['y_pixels_in_detector']
        return cls(
            int(nimages[()]),
            int(ntrigger[()]),
            float(photon_energy[()]),
            int(x_pixels_in_detector[()]),
            int(y_pixels_in_detector[()]),
        )


@dataclass(frozen=True)
class DetectorGroup:
    detector_specific: DetectorSpecificGroup
    detector_distance_m: float
    beam_center_x_px: int
    beam_center_y_px: int
    bit_depth_readout: int
    x_pixel_size_m: float
    y_pixel_size_m: float

    @classmethod
    def read(cls, group: h5py.Group) -> DetectorGroup:
        detector_specific = DetectorSpecificGroup.read(group['detectorSpecific'])
        h5_detector_distance = group['detector_distance']
        assert h5_detector_distance.attrs['units'] == b'm'
        h5_beam_center_x = group['beam_center_x']
        assert h5_beam_center_x.attrs['units'] == b'pixel'
        h5_beam_center_y = group['beam_center_y']
        assert h5_beam_center_y.attrs['units'] == b'pixel'
        h5_bit_depth_readout = group['bit_depth_readout']
        h5_x_pixel_size = group['x_pixel_size']
        assert h5_x_pixel_size.attrs['units'] == b'm'
        h5_y_pixel_size = group['y_pixel_size']
        assert h5_y_pixel_size.attrs['units'] == b'm'
        return cls(
            detector_specific,
            float(h5_detector_distance[()]),
            int(h5_beam_center_x[()]),
            int(h5_beam_center_y[()]),
            int(h5_bit_depth_readout[()]),
            float(h5_x_pixel_size[()]),
            float(h5_y_pixel_size[()]),
        )


@dataclass(frozen=True)
class InstrumentGroup:
    detector: DetectorGroup

    @classmethod
    def read(cls, group: h5py.Group) -> InstrumentGroup:
        detector = DetectorGroup.read(group['detector'])
        return cls(detector)


@dataclass(frozen=True)
class GoniometerGroup:
    chi_deg: float

    @classmethod
    def read(cls, group: h5py.Group) -> GoniometerGroup:
        chi_item = group['chi']
        chi_space = chi_item.id.get_space()

        assert chi_item.attrs['units'] == b'degree'

        if chi_space.get_simple_extent_type() == h5py.h5s.SCALAR:
            chi_deg = float(chi_item[()])
        elif isinstance(chi_item, h5py.Dataset):
            chi_deg = float(chi_item[0])
        else:
            raise ValueError('Failed to read goniometer angle (chi)!')

        return cls(chi_deg)


@dataclass(frozen=True)
class SampleGroup:
    goniometer: GoniometerGroup

    @classmethod
    def read(cls, group: h5py.Group) -> SampleGroup:
        goniometer = GoniometerGroup.read(group['goniometer'])
        return cls(goniometer)


@dataclass(frozen=True)
class EntryGroup:
    data: DataGroup
    instrument: InstrumentGroup
    sample: SampleGroup

    @classmethod
    def read(cls, group: h5py.Group, num_patterns_per_array: int) -> EntryGroup:
        data = DataGroup.read(group['data'], num_patterns_per_array)
        instrument = InstrumentGroup.read(group['instrument'])
        sample = SampleGroup.read(group['sample'])
        return cls(data, instrument, sample)


class VelociprobeDiffractionDataset(DiffractionDataset):
    def __init__(
        self,
        metadata: DiffractionMetadata,
        contents_tree: SimpleTreeNode,
        entry: EntryGroup,
    ) -> None:
        self._metadata = metadata
        self._contents_tree = contents_tree
        self._entry = entry

    def get_metadata(self) -> DiffractionMetadata:
        return self._metadata

    def get_layout(self) -> SimpleTreeNode:
        return self._contents_tree

    def get_bad_pixels(self) -> BadPixels | None:
        return None

    @overload
    def __getitem__(self, index: int) -> DiffractionArray: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[DiffractionArray]: ...

    def __getitem__(self, index: int | slice) -> DiffractionArray | Sequence[DiffractionArray]:
        return self._entry.data[index]

    def __len__(self) -> int:
        return len(self._entry.data)


class VelociprobeDiffractionFileReader(DiffractionFileReader):
    def __init__(self) -> None:
        super().__init__()
        self._tree_builder = H5DiffractionFileTreeBuilder()
        self.stage_rotation_deg = 0.0  # TODO This is a hack; remove when able!

    def read(self, file_path: Path) -> DiffractionDataset:
        dataset: DiffractionDataset = SimpleDiffractionDataset.create_null(file_path)

        with h5py.File(file_path, 'r') as h5_file:
            h5_dataset = h5_file['/entry/data/data_000001']
            num_patterns_per_array = h5_dataset.shape[0]
            pattern_dtype = h5_dataset.dtype

            entry = EntryGroup.read(h5_file['entry'], num_patterns_per_array)
            detector = entry.instrument.detector
            detector_pixel_geometry = PixelGeometry(
                detector.x_pixel_size_m,
                detector.y_pixel_size_m,
            )
            crop_center = CropCenter(
                detector.beam_center_x_px,
                detector.beam_center_y_px,
            )
            detector_specific = detector.detector_specific
            detector_extent = ImageExtent(
                detector_specific.x_pixels_in_detector,
                detector_specific.y_pixels_in_detector,
            )
            probe_energy_eV = detector_specific.photon_energy_eV  # noqa: N806
            num_arrays = detector_specific.num_patterns_total // num_patterns_per_array

            metadata = DiffractionMetadata(
                num_patterns_per_array=[num_patterns_per_array] * num_arrays,
                pattern_dtype=pattern_dtype,
                detector_distance_m=detector.detector_distance_m,
                detector_extent=detector_extent,
                detector_pixel_geometry=detector_pixel_geometry,
                detector_bit_depth=detector.bit_depth_readout,
                crop_center=crop_center,
                probe_energy_eV=probe_energy_eV,
                file_path=file_path,
            )
            contents_tree = self._tree_builder.build(h5_file)
            dataset = VelociprobeDiffractionDataset(metadata, contents_tree, entry)

            # vvv TODO This is a hack; remove when able! vvv
            self.stage_rotation_deg = entry.sample.goniometer.chi_deg

        return dataset



from pathlib import Path

import numpy

from ptychodus.api.object import Object, ObjectFileReader, ObjectFileWriter
from ptychodus.api.plugins import PluginRegistry


class NPYObjectFileReader(ObjectFileReader):
    def read(self, file_path: Path) -> Object:
        array = numpy.load(file_path)
        return Object(array=array, pixel_geometry=None, center=None)


class NPYObjectFileWriter(ObjectFileWriter):
    def write(self, file_path: Path, object_: Object) -> None:
        array = object_.get_array()
        numpy.save(file_path, array)


def register_plugins(registry: PluginRegistry) -> None:
    registry.object_file_readers.register_plugin(
        NPYObjectFileReader(),
        simple_name='NPY',
        display_name='NumPy Binary Files (*.npy)',
    )
    registry.object_file_writers.register_plugin(
        NPYObjectFileWriter(),
        simple_name='NPY',
        display_name='NumPy Binary Files (*.npy)',
    )



import logging

from PyQt5.QtCore import QRectF

from ptychodus.api.geometry import Box2D
from ptychodus.api.observer import Observable, Observer

from ...model.analysis import FourierAnalyzer
from ...model.visualization import VisualizationEngine
from ...view.object import FourierAnalysisDialog
from ...view.widgets import ExceptionDialog
from ..data import FileDialogFactory
from ..image import ImageController

logger = logging.getLogger(__name__)


class FourierAnalysisViewController(Observer):
    def __init__(
        self,
        analyzer: FourierAnalyzer,
        real_space_visualization_engine: VisualizationEngine,
        reciprocal_space_visualization_engine: VisualizationEngine,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._analyzer = analyzer
        self._dialog = FourierAnalysisDialog()
        self._dialog.setWindowTitle('Fourier Analysis')

        self._real_space_image_controller = ImageController(
            real_space_visualization_engine,
            self._dialog.real_space_view,
            self._dialog.status_bar,
            file_dialog_factory,
        )
        self._reciprocal_space_image_controller = ImageController(
            reciprocal_space_visualization_engine,
            self._dialog.reciprocal_space_view,
            self._dialog.status_bar,
            file_dialog_factory,
        )

        item_signals = self._real_space_image_controller.get_item().get_signals()
        item_signals.fourier_finished.connect(self._analyze_fourier)

        analyzer.add_observer(self)

    def analyze(self, item_index: int) -> None:
        self._analyzer.set_product(item_index)
        self._dialog.open()

    def _analyze_fourier(self, rect: QRectF) -> None:
        if rect.isEmpty():
            logger.debug('QRectF is empty!')
            return

        box = Box2D(
            x=rect.x(),
            y=rect.y(),
            width=rect.width(),
            height=rect.height(),
        )

        try:
            self._analyzer.analyze_roi(box)
        except ValueError as exc:
            logger.exception(exc)
            ExceptionDialog.show_exception('Fourier Analysis', exc)

    def _sync_model_to_view(self) -> None:
        try:
            object_ = self._analyzer.get_object()
        except Exception as exc:
            logger.exception(exc)
            ExceptionDialog.show_exception('Fourier Analysis', exc)
        else:
            self._real_space_image_controller.set_array(
                object_.get_layer(0), object_.get_pixel_geometry()
            )

        try:
            result = self._analyzer.get_result()
        except ValueError:
            self._reciprocal_space_image_controller.clear_array()
        except Exception as exc:
            logger.exception(exc)
            ExceptionDialog.show_exception('Fourier Analysis', exc)
        else:
            self._reciprocal_space_image_controller.set_array(
                result.transformed_roi, result.pixel_geometry
            )

    def _update(self, observable: Observable) -> None:
        if observable is self._analyzer:
            self._sync_model_to_view()



from typing import Any, TypeAlias

from scipy.integrate import quad
from scipy.special import fresnel, j0
import numpy
import numpy.typing

ComplexArrayType: TypeAlias = numpy.typing.NDArray[numpy.complexfloating[Any, Any]]
RealArrayType: TypeAlias = numpy.typing.NDArray[numpy.floating[Any]]


class SquareAperture:
    def __init__(self, width_m: float, wavelength_m: float) -> None:
        self._width_m = width_m
        self._wavelength_m = wavelength_m

    def get_fresnel_number(self, z_m: float) -> float:
        upper = (self._width_m / 2) ** 2
        lower = self._wavelength_m * z_m
        return upper / lower

    def _integral1d(self, r_m: RealArrayType, z_m: float) -> ComplexArrayType:
        sqrt2NF = numpy.sqrt(2 * self.get_fresnel_number(z_m))  # noqa: N806
        xi = 2 * r_m / self._width_m
        Sm, Cm = fresnel(sqrt2NF * (1 - xi))  # noqa: N806
        Sp, Cp = fresnel(sqrt2NF * (1 + xi))  # noqa: N806
        return (Cm + Cp + 1j * (Sm + Sp)) / numpy.sqrt(2)

    def diffract(self, x_m: RealArrayType, y_m: RealArrayType, z_m: float) -> ComplexArrayType:
        """Fresnel diffraction; see Goodman p.100"""
        assert x_m.shape == y_m.shape
        Ix = self._integral1d(x_m, z_m)  # noqa: N806
        Iy = self._integral1d(y_m, z_m)  # noqa: N806
        return Ix * Iy * numpy.exp(2j * numpy.pi * z_m / self._wavelength_m) / 1j


class CircularAperture:
    def __init__(self, diameter_m: float, wavelength_m: float) -> None:
        self._diameter_m = diameter_m
        self._wavelength_m = wavelength_m

    def get_fresnel_number(self, z_m: float) -> float:
        upper = (self._diameter_m / 2) ** 2
        lower = self._wavelength_m * z_m
        return upper / lower

    def diffract(self, x_m: RealArrayType, y_m: RealArrayType, z_m: float) -> ComplexArrayType:
        """Fresnel diffraction; see Goodman p.102"""
        assert x_m.shape == y_m.shape

        twopi = 2 * numpy.pi
        sqrtLZ = numpy.sqrt(self._wavelength_m * z_m)  # noqa: N806
        sqrtNF = numpy.sqrt(self.get_fresnel_number(z_m))  # noqa: N806

        rhop = rho / sqrtLZ
        rp = numpy.hypot(x_m, y_m) / sqrtLZ

        arg = twopi * rhop * rp
        integrand = rhop * numpy.exp(1j * numpy.pi * rhop**2) * j0(arg)
        integral = integrate(integrand, 0, sqrtNF)
        return twopi * integral



from .api import DiffractionAPI, PatternsStreamingContext
from .bad_pixels import BadPixelsProvider
from .core import DiffractionCore
from .dataset import (
    AssembledDiffractionArray,
    AssembledDiffractionDataset,
    DiffractionDatasetObserver,
)
from .settings import DetectorSettings, DiffractionSettings
from .sizer import PatternSizer

__all__ = [
    'AssembledDiffractionArray',
    'AssembledDiffractionDataset',
    'BadPixelsProvider',
    'DetectorSettings',
    'DiffractionAPI',
    'DiffractionCore',
    'DiffractionDatasetObserver',
    'DiffractionSettings',
    'PatternSizer',
    'PatternsStreamingContext',
]



from collections.abc import Callable, Iterable, Iterator
from pathlib import Path
import logging

from ptychodus.api.plugins import PluginChooser
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePositionFileWriter,
)

from .builder import FromFileProbePositionsBuilder, ProbePositionsBuilder
from .cartesian import CartesianProbePositionsBuilder, CartesianProbePositionsVariant
from .concentric import ConcentricProbePositionsBuilder
from .lissajous import LissajousProbePositionsBuilder
from .settings import ProbePositionsSettings
from .spiral import SpiralProbePositionsBuilder

logger = logging.getLogger(__name__)


class ProbePositionsBuilderFactory(Iterable[str]):
    def __init__(
        self,
        settings: ProbePositionsSettings,
        file_reader_chooser: PluginChooser[ProbePositionFileReader],
        file_writer_chooser: PluginChooser[ProbePositionFileWriter],
    ) -> None:
        self._settings = settings
        self._file_reader_chooser = file_reader_chooser
        self._file_writer_chooser = file_writer_chooser
        self._builders: dict[str, Callable[[], ProbePositionsBuilder]] = {
            variant.name.lower(): lambda var=variant: CartesianProbePositionsBuilder(var, settings)  # type: ignore
            for variant in CartesianProbePositionsVariant
        }
        self._builders.update(
            {
                'concentric': lambda: ConcentricProbePositionsBuilder(settings),
                'spiral': lambda: SpiralProbePositionsBuilder(settings),
                'lissajous': lambda: LissajousProbePositionsBuilder(settings),
            }
        )

    def __iter__(self) -> Iterator[str]:
        return iter(self._builders)

    def create(self, name: str) -> ProbePositionsBuilder:
        try:
            factory = self._builders[name]
        except KeyError as exc:
            raise KeyError(f'Unknown scan builder "{name}"!') from exc

        return factory()

    def create_default(self) -> ProbePositionsBuilder:
        return next(iter(self._builders.values()))()

    def create_from_settings(self) -> ProbePositionsBuilder:
        name = self._settings.builder.get_value()
        name_repaired = name.casefold()

        if name_repaired == 'from_file':
            return self.create_scan_from_file(
                self._settings.file_path.get_value(),
                self._settings.file_type.get_value(),
            )

        return self.create(name_repaired)

    def get_open_file_filters(self) -> Iterator[str]:
        for plugin in self._file_reader_chooser:
            yield plugin.display_name

    def get_open_file_filter(self) -> str:
        return self._file_reader_chooser.get_current_plugin().display_name

    def create_scan_from_file(self, file_path: Path, file_type: str) -> ProbePositionsBuilder:
        self._file_reader_chooser.set_current_plugin(file_type)
        file_type = self._file_reader_chooser.get_current_plugin().simple_name
        file_reader = self._file_reader_chooser.get_current_plugin().strategy
        return FromFileProbePositionsBuilder(self._settings, file_path, file_type, file_reader)

    def get_save_file_filters(self) -> Iterator[str]:
        for plugin in self._file_writer_chooser:
            yield plugin.display_name

    def get_save_file_filter(self) -> str:
        return self._file_writer_chooser.get_current_plugin().display_name

    def save_scan(self, file_path: Path, file_type: str, scan: ProbePositionSequence) -> None:
        self._file_writer_chooser.set_current_plugin(file_type)
        file_type = self._file_writer_chooser.get_current_plugin().simple_name
        logger.debug(f'Writing "{file_path}" as "{file_type}"')
        file_writer = self._file_writer_chooser.get_current_plugin().strategy
        file_writer.write(file_path, scan)



from pathlib import Path
from typing import Any, Final
import logging

import numpy

from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.object import Object, ObjectCenter
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe import ProbeSequence
from ptychodus.api.product import (
    Product,
    ProductFileReader,
    ProductFileWriter,
    ProductMetadata,
)
from ptychodus.api.reconstructor import LossValue
from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition

logger = logging.getLogger(__name__)


class NPZProductFileIO(ProductFileReader, ProductFileWriter):
    SIMPLE_NAME: Final[str] = 'NPZ'
    DISPLAY_NAME: Final[str] = 'Ptychodus NumPy Zipped Archive (*.npz)'

    NAME: Final[str] = 'name'
    COMMENTS: Final[str] = 'comments'
    DETECTOR_OBJECT_DISTANCE: Final[str] = 'detector_object_distance_m'
    PROBE_ENERGY: Final[str] = 'probe_energy_eV'
    PROBE_PHOTON_COUNT: Final[str] = 'probe_photon_count'
    EXPOSURE_TIME: Final[str] = 'exposure_time_s'
    MASS_ATTENUATION: Final[str] = 'mass_attenuation_m2_kg'
    TOMOGRAPHY_ANGLE: Final[str] = 'tomography_angle_deg'

    PROBE_ARRAY: Final[str] = 'probe'
    OPR_WEIGHTS: Final[str] = 'opr_weights'
    PROBE_PIXEL_HEIGHT: Final[str] = 'probe_pixel_height_m'
    PROBE_PIXEL_WIDTH: Final[str] = 'probe_pixel_width_m'
    PROBE_POSITION_INDEXES: Final[str] = 'probe_position_indexes'
    PROBE_POSITION_X: Final[str] = 'probe_position_x_m'
    PROBE_POSITION_Y: Final[str] = 'probe_position_y_m'

    OBJECT_ARRAY: Final[str] = 'object'
    OBJECT_CENTER_X: Final[str] = 'object_center_x_m'
    OBJECT_CENTER_Y: Final[str] = 'object_center_y_m'
    OBJECT_LAYER_SPACING: Final[str] = 'object_layer_spacing_m'
    OBJECT_PIXEL_HEIGHT: Final[str] = 'object_pixel_height_m'
    OBJECT_PIXEL_WIDTH: Final[str] = 'object_pixel_width_m'

    LOSS_EPOCHS: Final[str] = 'loss_epochs'
    LOSS_VALUES: Final[str] = 'loss_values'

    def read(self, file_path: Path) -> Product:
        with numpy.load(file_path) as npz_file:
            probe_photon_count = 0.0

            try:
                probe_photon_count = float(npz_file[self.PROBE_PHOTON_COUNT])
            except KeyError:
                logger.debug('Probe photon count not found.')

            mass_attenuation_m2_kg = 0.0

            try:
                mass_attenuation_m2_kg = float(npz_file[self.MASS_ATTENUATION])
            except KeyError:
                logger.debug('Mass attenuation not found.')

            tomography_angle_deg = 0.0

            try:
                tomography_angle_deg = float(npz_file[self.TOMOGRAPHY_ANGLE])
            except KeyError:
                logger.debug('Tomography angle not found.')

            metadata = ProductMetadata(
                name=str(npz_file[self.NAME]),
                comments=str(npz_file[self.COMMENTS]),
                detector_distance_m=float(npz_file[self.DETECTOR_OBJECT_DISTANCE]),
                probe_energy_eV=float(npz_file[self.PROBE_ENERGY]),
                probe_photon_count=probe_photon_count,
                exposure_time_s=float(npz_file[self.EXPOSURE_TIME]),
                mass_attenuation_m2_kg=mass_attenuation_m2_kg,
                tomography_angle_deg=tomography_angle_deg,
            )

            scan_indexes = npz_file[self.PROBE_POSITION_INDEXES]
            scan_x_m = npz_file[self.PROBE_POSITION_X]
            scan_y_m = npz_file[self.PROBE_POSITION_Y]

            probe_pixel_geometry = PixelGeometry(
                width_m=float(npz_file[self.PROBE_PIXEL_WIDTH]),
                height_m=float(npz_file[self.PROBE_PIXEL_HEIGHT]),
            )

            try:
                opr_weights = npz_file[self.OPR_WEIGHTS]
            except KeyError:
                logger.debug('OPR weights not found.')
                opr_weights = None

            probe = ProbeSequence(
                array=npz_file[self.PROBE_ARRAY],
                opr_weights=opr_weights,
                pixel_geometry=probe_pixel_geometry,
            )

            object_pixel_geometry = PixelGeometry(
                width_m=float(npz_file[self.OBJECT_PIXEL_WIDTH]),
                height_m=float(npz_file[self.OBJECT_PIXEL_HEIGHT]),
            )
            object_center = ObjectCenter(
                coordinate_x_m=float(npz_file[self.OBJECT_CENTER_X]),
                coordinate_y_m=float(npz_file[self.OBJECT_CENTER_Y]),
            )
            object_ = Object(
                array=npz_file[self.OBJECT_ARRAY],
                pixel_geometry=object_pixel_geometry,
                center=object_center,
                layer_spacing_m=npz_file[self.OBJECT_LAYER_SPACING],
            )

            try:
                loss_values = npz_file[self.LOSS_VALUES]
            except KeyError:
                loss_values = npz_file['costs']

            try:
                loss_epochs = npz_file[self.LOSS_EPOCHS]
            except KeyError:
                loss_epochs = numpy.arange(len(loss_values))

        point_list: list[ProbePosition] = []

        for idx, x_m, y_m in zip(scan_indexes, scan_x_m, scan_y_m):
            point = ProbePosition(idx, x_m, y_m)
            point_list.append(point)

        losses: list[LossValue] = []

        for epoch, value in zip(loss_epochs, loss_values):
            loss = LossValue(epoch, value)
            losses.append(loss)

        return Product(
            metadata=metadata,
            probe_positions=ProbePositionSequence(point_list),
            probes=probe,
            object_=object_,
            losses=losses,
        )

    def write(self, file_path: Path, product: Product) -> None:
        contents: dict[str, Any] = dict()
        scan_indexes: list[int] = []
        scan_x_m: list[float] = []
        scan_y_m: list[float] = []

        for point in product.probe_positions:
            scan_indexes.append(point.index)
            scan_x_m.append(point.coordinate_x_m)
            scan_y_m.append(point.coordinate_y_m)

        metadata = product.metadata
        contents[self.NAME] = metadata.name
        contents[self.COMMENTS] = metadata.comments
        contents[self.DETECTOR_OBJECT_DISTANCE] = metadata.detector_distance_m
        contents[self.PROBE_ENERGY] = metadata.probe_energy_eV
        contents[self.PROBE_PHOTON_COUNT] = metadata.probe_photon_count
        contents[self.EXPOSURE_TIME] = metadata.exposure_time_s
        contents[self.MASS_ATTENUATION] = metadata.mass_attenuation_m2_kg

        contents[self.PROBE_POSITION_INDEXES] = scan_indexes
        contents[self.PROBE_POSITION_X] = scan_x_m
        contents[self.PROBE_POSITION_Y] = scan_y_m

        probe = product.probes
        contents[self.PROBE_ARRAY] = probe.get_array()

        try:
            opr_weights = probe.get_opr_weights()
        except ValueError:
            pass
        else:
            contents[self.OPR_WEIGHTS] = opr_weights

        probe_pixel_geometry = probe.get_pixel_geometry()
        contents[self.PROBE_PIXEL_WIDTH] = probe_pixel_geometry.width_m
        contents[self.PROBE_PIXEL_HEIGHT] = probe_pixel_geometry.height_m

        object_ = product.object_
        object_geometry = object_.get_geometry()
        contents[self.OBJECT_ARRAY] = object_.get_array()
        contents[self.OBJECT_CENTER_X] = object_geometry.center_x_m
        contents[self.OBJECT_CENTER_Y] = object_geometry.center_y_m
        contents[self.OBJECT_PIXEL_WIDTH] = object_geometry.pixel_width_m
        contents[self.OBJECT_PIXEL_HEIGHT] = object_geometry.pixel_height_m
        contents[self.OBJECT_LAYER_SPACING] = object_.layer_spacing_m

        loss_epochs: list[int] = []
        loss_values: list[float] = []

        for loss in product.losses:
            loss_epochs.append(loss.epoch)
            loss_values.append(loss.value)

        contents[self.LOSS_EPOCHS] = loss_epochs
        contents[self.LOSS_VALUES] = loss_values

        numpy.savez(file_path, **contents)


def register_plugins(registry: PluginRegistry) -> None:
    npz_product_file_io = NPZProductFileIO()

    registry.register_product_file_reader_with_adapters(
        npz_product_file_io,
        simple_name=NPZProductFileIO.SIMPLE_NAME,
        display_name=NPZProductFileIO.DISPLAY_NAME,
    )
    registry.product_file_writers.register_plugin(
        npz_product_file_io,
        simple_name=NPZProductFileIO.SIMPLE_NAME,
        display_name=NPZProductFileIO.DISPLAY_NAME,
    )



from .chat import ChatMessage, ChatHistory, ChatObserver, ChatRole
from .core import AgentCore, AgentPresenter
from .settings import ArgoSettings

__all__ = [
    'AgentCore',
    'AgentPresenter',
    'ArgoSettings',
    'ChatHistory',
    'ChatMessage',
    'ChatObserver',
    'ChatRole',
]



"""Simple tree data structure used for representing dataset layouts."""

from __future__ import annotations
from collections.abc import Sequence


class SimpleTreeNode:
    """Node in a string-labelled tree, used to expose dataset hierarchy."""

    def __init__(self, parent_item: SimpleTreeNode | None, item_data: Sequence[str]) -> None:
        self.parent_item = parent_item
        self.item_data = item_data
        self.child_items: list[SimpleTreeNode] = list()

    @classmethod
    def create_root(cls, item_data: Sequence[str]) -> SimpleTreeNode:
        return cls(None, item_data)

    def create_child(self, item_data: Sequence[str]) -> SimpleTreeNode:
        child_item = SimpleTreeNode(self, item_data)
        self.child_items.append(child_item)
        return child_item

    @property
    def is_root(self) -> bool:
        return self.parent_item is None

    @property
    def is_leaf(self) -> bool:
        return not self.child_items

    def data(self, column: int) -> str | None:
        try:
            return self.item_data[column]
        except IndexError:
            return None

    def row(self) -> int:
        if self.parent_item:
            return self.parent_item.child_items.index(self)

        return 0



from __future__ import annotations
import logging

from ptychodus.api.observer import Observable
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe import ProbeSequence, ProbeGeometryProvider

from .builder import FromMemoryProbeBuilder, ProbeSequenceBuilder
from .multimodal import MultimodalProbeBuilder
from .settings import ProbeSettings

logger = logging.getLogger(__name__)


class ProbeRepositoryItem(ParameterGroup):
    def __init__(
        self,
        geometry_provider: ProbeGeometryProvider,
        settings: ProbeSettings,
        builder: ProbeSequenceBuilder,
        additional_modes_builder: MultimodalProbeBuilder,
    ) -> None:
        super().__init__()
        self._geometry_provider = geometry_provider
        self._settings = settings
        self._builder = builder
        self._additional_modes_builder = additional_modes_builder
        self._probe_seq = ProbeSequence(array=None, opr_weights=None, pixel_geometry=None)

        self._add_group('builder', builder, observe=True)
        self._add_group('additional_modes', additional_modes_builder, observe=True)

        self._rebuild()

    def assign_item(self, item: ProbeRepositoryItem) -> None:
        group = 'additional_modes'

        self._remove_group(group)
        self._additional_modes_builder.remove_observer(self)

        additional_modes_builder = item.get_additional_modes_builder()

        self._additional_modes_builder = additional_modes_builder.copy()
        self._additional_modes_builder.add_observer(self)
        self._add_group(group, self._additional_modes_builder, observe=True)

        self.set_builder(item.get_builder().copy())
        self._rebuild()

    def assign(self, probe: ProbeSequence) -> None:
        builder = FromMemoryProbeBuilder(self._settings, probe)
        self.set_builder(builder)

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

        self._builder.sync_to_settings()
        self._additional_modes_builder.sync_to_settings()

    def get_probes(self) -> ProbeSequence:
        return self._probe_seq

    def get_builder(self) -> ProbeSequenceBuilder:
        return self._builder

    def set_builder(self, builder: ProbeSequenceBuilder) -> None:
        group = 'builder'
        self._remove_group(group)
        self._builder.remove_observer(self)
        self._builder = builder
        self._builder.add_observer(self)
        self._add_group(group, self._builder, observe=True)
        self._rebuild()

    def _rebuild(self) -> None:
        try:
            probe = self._builder.build(self._geometry_provider)
        except Exception:
            logger.exception('Failed to rebuild probe!')
            return

        self._probe_seq = self._additional_modes_builder.build(probe, self._geometry_provider)
        self.notify_observers()

    def get_additional_modes_builder(self) -> MultimodalProbeBuilder:
        return self._additional_modes_builder

    def _update(self, observable: Observable) -> None:
        if observable is self._builder:
            self._rebuild()
        elif observable is self._additional_modes_builder:
            self._rebuild()
        else:
            super()._update(observable)



from pathlib import Path
from typing import Any
import logging
import threading
import time

import numpy

from pvapy.hpc.adImageProcessor import AdImageProcessor
from pvapy.utility.floatWithUnits import FloatWithUnits
from pvapy.utility.timeUtility import TimeUtility
import pvaccess
import pvapy

from ptychodus.model import ModelCore
import ptychodus
import ptychodus.api


class ReconstructionThread(threading.Thread):
    def __init__(
        self,
        ptychodus: ModelCore,
        input_product_path: Path,
        output_product_path: Path,
        reconstruct_pv: str,
    ) -> None:
        super().__init__()
        self._ptychodus = ptychodus
        self._input_product_path = input_product_path
        self._output_product_path = output_product_path
        self._channel = pvapy.Channel(reconstruct_pv, pvapy.CA)
        self._reconstruct_event = threading.Event()
        self._stop_event = threading.Event()

        self._ptychodus_streaming_context = None

        self._channel.subscribe('reconstructor', self._monitor)
        self._channel.startMonitor()

    def run(self) -> None:
        while not self._stop_event.is_set():
            if self._reconstruct_event.wait(timeout=1.0):
                logging.debug('ReconstructionThread: Begin assembling scan positions')

                if self._ptychodus_streaming_context is not None:
                    self._ptychodus_streaming_context.stop()

                logging.debug('ReconstructionThread: End assembling scan positions')
                self._ptychodus.batch_mode_execute(
                    'reconstruct', self._input_product_path, self._output_product_path
                )
                self._reconstruct_event.clear()
                # reconstruction done; indicate that results are ready
                self._channel.put(0)

    def _monitor(self, pv_object: pvaccess.PvObject) -> None:
        # NOTE caput bdpgp:gp:bit3 1
        logging.debug(f'ReconstructionThread::monitor {pv_object}')

        if pv_object['value']['index'] == 1:
            logging.debug('ReconstructionThread: Reconstruct PV triggered!')
            # start reconstructing
            self._reconstruct_event.set()
        else:
            logging.debug('ReconstructionThread: Reconstruct PV not triggered!')

    def stop(self) -> None:
        self._stop_event.set()


class PtychodusAdImageProcessor(AdImageProcessor):
    def __init__(self, config_dict: dict[str, Any] = {}) -> None:
        super().__init__(config_dict)

        self.logger.debug(f'{ptychodus.__name__.title()} ({ptychodus.__version__})')

        settings_file = config_dict.get('settingsFile')
        self._ptychodus = ModelCore(settings_file)
        self._reconstruction_thread = ReconstructionThread(
            self._ptychodus,
            Path(config_dict.get('inputProductPath', 'input.npz')),
            Path(config_dict.get('outputProductPath', 'output.npz')),
            config_dict.get('reconstructPV', 'bdpgp:gp:bit3'),
        )
        self._pos_x_pv = config_dict.get('pos_x_pv', 'bluesky:pos_x')
        self._pos_y_pv = config_dict.get('pos_y_pv', 'bluesky:pos_y')
        self._num_frames_processed = 0
        self._processing_time = 0.0

    def start(self) -> None:
        """Called at startup"""
        self._ptychodus.__enter__()
        self._reconstruction_thread.start()

    def stop(self) -> None:
        """Called at shutdown"""
        self._reconstruction_thread.stop()
        self._reconstruction_thread.join()
        self._ptychodus.__exit__(None, None, None)

    def configure(self, config_dict: dict[str, Any]) -> None:
        """Configures user processor"""
        num_arrays = config_dict['num_arrays']
        num_patterns_per_array = config_dict.get('num_patterns_per_array', 1)
        pattern_dtype = config_dict.get('pattern_dtype', 'uint16')

        metadata = ptychodus.api.diffraction.DiffractionMetadata(
            num_patterns_per_array=[int(num_patterns_per_array)] * int(num_arrays),
            pattern_dtype=numpy.dtype(pattern_dtype),
        )
        self._ptychodus_streaming_context = self._ptychodus.create_streaming_context(metadata)
        self._ptychodus_streaming_context.start()  # TODO clean up

    def process(self, pv_object: pvaccess.PvObject) -> pvaccess.PvObject:
        """Processes monitor update"""
        processing_begin_time = time.time()

        (frame_id, image, nx, ny, nz, color_mode, field_key) = self.reshapeNtNdArray(pv_object)
        frame_time_stamp = TimeUtility.getTimeStampAsFloat(pv_object['timeStamp'])

        if nx is None:
            self.logger.debug(f'Frame id {frame_id} contains an empty image.')
        else:
            self.logger.debug(f'Frame id {frame_id} time stamp {frame_time_stamp}')
            image3d = image[numpy.newaxis, :, :].copy()
            array = ptychodus.api.diffraction.SimpleDiffractionArray(
                label=f'Frame{frame_id}',
                indexes=numpy.array([frame_id]),
                patterns=image3d,
            )
            self._ptychodus_streaming_context.append_array(array)

        pos_x_queue = self.metadataQueueMap[self._pos_x_pv]

        while True:
            try:
                pos_x = pos_x_queue.get(0)
            except pvaccess.QueueEmpty:
                break
            else:
                self._ptychodus_streaming_context.append_positions_x(
                    pos_x['values'],
                    [TimeUtility.getTimeStampAsFloat(ts) for ts in pos_x['t']],
                )

        pos_y_queue = self.metadataQueueMap[self._pos_y_pv]

        while True:
            try:
                pos_y = pos_y_queue.get(0)
            except pvaccess.QueueEmpty:
                break
            else:
                self._ptychodus_streaming_context.append_positions_y(
                    pos_y['values'],
                    [TimeUtility.getTimeStampAsFloat(ts) for ts in pos_y['t']],
                )

        processing_end_time = time.time()
        self._processing_time += processing_end_time - processing_begin_time
        self._num_frames_processed += 1

        return pv_object

    def resetStats(self) -> None:  # noqa: N802
        """Resets statistics for user processor"""
        self._num_frames_processed = 0
        self._processing_time = 0.0

    def getStats(self) -> dict[str, Any]:  # noqa: N802
        """Retrieves statistics for user processor"""
        num_frames_queued = self._ptychodus_streaming_context.get_queue_size()
        processed_frame_rate = 0.0

        if self._processing_time > 0.0:
            processed_frame_rate = self._num_frames_processed / self._processing_time

        return {
            'num_frames_processed': self._num_frames_processed,
            'num_frames_queued': num_frames_queued,
            'processing_time': FloatWithUnits(self._processing_time, 's'),
            'processed_frame_rate': FloatWithUnits(processed_frame_rate, 'fps'),
        }

    def getStatsPvaTypes(self) -> dict[str, pvaccess.ScalarType]:  # noqa: N802
        """Defines PVA types for different stats variables"""
        return {
            'num_frames_processed': pvaccess.UINT,
            'num_frames_queued': pvaccess.UINT,
            'processing_time': pvaccess.DOUBLE,
            'processing_frame_rate': pvaccess.DOUBLE,
        }



import os

import numpy as np
import matplotlib.pyplot as plt

import ptychodus.model.phase_unwrapper as pu


def test_phase_unwrap() -> None:
    phase_unwrapper = pu.PhaseUnwrapper(
        image_grad_method='fourier_differentiation',
        image_integration_method='fourier',
    )
    img = np.load(os.path.join('data', 'phase_unwrap', 'recon_20241220_epoch_400.npy'))
    img = img[0]

    phase = phase_unwrapper.unwrap(img)

    plt.figure()
    plt.imshow(phase)
    plt.show()


if __name__ == '__main__':
    test_phase_unwrap()



from collections.abc import Sequence
from dataclasses import dataclass


@dataclass(frozen=True)
class ChatModelProperties:
    name: str

    @property
    def is_o_series_model(self) -> bool:
        return self.name.startswith('gpto')

    @property
    def accepts_system_prompt(self) -> bool:
        return not self.is_o_series_model

    @property
    def accepts_stop_sequence(self) -> bool:
        return not self.is_o_series_model

    @property
    def accepts_temperature(self) -> bool:
        return not self.is_o_series_model

    @property
    def accepts_top_p(self) -> bool:
        return not self.is_o_series_model

    @property
    def accepts_max_tokens(self) -> bool:
        return not self.is_o_series_model

    @property
    def accepts_max_completion_tokens(self) -> bool:
        return self.is_o_series_model


def list_argo_model_properties() -> Sequence[ChatModelProperties]:
    return [
        ChatModelProperties(name='gpt35'),
        ChatModelProperties(name='gpt35large'),
        ChatModelProperties(name='gpt4'),
        ChatModelProperties(name='gpt4large'),
        ChatModelProperties(name='gpt4o'),
        ChatModelProperties(name='gpt4olatest'),
        ChatModelProperties(name='gpt4turbo'),
        ChatModelProperties(name='gpto1'),
        ChatModelProperties(name='gpto1mini'),
        ChatModelProperties(name='gpto3mini'),
    ]



from __future__ import annotations
from collections.abc import Mapping
from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from typing import Any, Generic, TypeVar
import logging
import sys
import typing
import yaml

try:
    # xdrlib removed from the standard library in Python 3.13
    import xdrlib  # type: ignore[import-not-found]
except ModuleNotFoundError:
    # use full module path to make this file usable as an entry point
    from ptychodus.plugins._xdrlib import xdrlib  # type: ignore[no-redef]

import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
)

T = TypeVar('T')

logger = logging.getLogger(__name__)


class EpicsType(IntEnum):
    DBR_STRING = 0
    DBR_SHORT = 1
    DBR_FLOAT = 2
    DBR_ENUM = 3
    DBR_CHAR = 4
    DBR_LONG = 5
    DBR_DOUBLE = 6
    DBR_STS_STRING = 7
    DBR_STS_SHORT = 8
    DBR_STS_FLOAT = 9
    DBR_STS_ENUM = 10
    DBR_STS_CHAR = 11
    DBR_STS_LONG = 12
    DBR_STS_DOUBLE = 13
    DBR_TIME_STRING = 14
    DBR_TIME_SHORT = 15
    DBR_TIME_FLOAT = 16
    DBR_TIME_ENUM = 17
    DBR_TIME_CHAR = 18
    DBR_TIME_LONG = 19
    DBR_TIME_DOUBLE = 20
    DBR_GR_STRING = 21
    DBR_GR_SHORT = 22
    DBR_GR_FLOAT = 23
    DBR_GR_ENUM = 24
    DBR_GR_CHAR = 25
    DBR_GR_LONG = 26
    DBR_GR_DOUBLE = 27
    DBR_CTRL_STRING = 28
    DBR_CTRL_SHORT = 29
    DBR_CTRL_FLOAT = 30
    DBR_CTRL_ENUM = 31
    DBR_CTRL_CHAR = 32
    DBR_CTRL_LONG = 33
    DBR_CTRL_DOUBLE = 34


def read_int_from_buffer(fp: typing.BinaryIO) -> int:
    unpacker = xdrlib.Unpacker(fp.read(4))
    return unpacker.unpack_int()


def read_float_from_buffer(fp: typing.BinaryIO) -> float:
    unpacker = xdrlib.Unpacker(fp.read(4))
    return unpacker.unpack_float()


def read_counted_string(unpacker: xdrlib.Unpacker) -> str:
    length = unpacker.unpack_int()
    return unpacker.unpack_string().decode() if length else str()


def read_counted_string_from_buffer(fp: typing.BinaryIO) -> str:
    length = read_int_from_buffer(fp)

    if length:
        sz = (length + 3) // 4 * 4 + 4
        unpacker = xdrlib.Unpacker(fp.read(sz))
        return unpacker.unpack_string().decode()

    return str()


@dataclass(frozen=True)
class MDAHeader:
    version: float
    scan_number: int
    dimensions: list[int]
    is_regular: bool
    extra_pvs_offset: int

    @classmethod
    def read(cls, fp: typing.BinaryIO) -> MDAHeader:
        unpacker = xdrlib.Unpacker(fp.read(12))
        version = unpacker.unpack_float()
        scan_number = unpacker.unpack_int()
        data_rank = unpacker.unpack_int()

        unpacker.reset(fp.read(4 * data_rank + 8))
        dimensions = unpacker.unpack_farray(data_rank, unpacker.unpack_int)
        is_regular = unpacker.unpack_bool()
        extra_pvs_offset = unpacker.unpack_int()

        return cls(version, scan_number, dimensions, is_regular, extra_pvs_offset)

    @property
    def data_rank(self) -> int:
        return len(self.dimensions)

    @property
    def has_extra_pvs(self) -> bool:
        return self.extra_pvs_offset > 0

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'version': self.version,
            'scan_number': self.scan_number,
            'dimensions': self.dimensions,
            'is_regular': self.is_regular,
            'extra_pvs_offset': self.extra_pvs_offset,
        }


@dataclass(frozen=True)
class MDAScanHeader:
    rank: int
    num_requested_points: int
    current_point: int
    lower_scan_offsets: list[int]

    @classmethod
    def read(cls, fp: typing.BinaryIO) -> MDAScanHeader:
        unpacker = xdrlib.Unpacker(fp.read(12))
        rank = unpacker.unpack_int()
        npts = unpacker.unpack_int()
        cpt = unpacker.unpack_int()
        lower_scan_offsets: list[int] = list()

        if rank > 1:
            unpacker.reset(fp.read(4 * npts))
            lower_scan_offsets = unpacker.unpack_farray(npts, unpacker.unpack_int)

        return cls(rank, npts, cpt, lower_scan_offsets)

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'rank': self.rank,
            'num_requested_points': self.num_requested_points,
            'current_point': self.current_point,
            'lower_scan_offsets': self.lower_scan_offsets,
        }


@dataclass(frozen=True)
class MDAScanPositionerInfo:
    number: int
    name: str
    description: str
    step_mode: str
    unit: str
    readback_name: str
    readback_description: str
    readback_unit: str

    @classmethod
    def read(cls, fp: typing.BinaryIO) -> MDAScanPositionerInfo:
        number = read_int_from_buffer(fp)
        name = read_counted_string_from_buffer(fp)
        description = read_counted_string_from_buffer(fp)
        step_mode = read_counted_string_from_buffer(fp)
        unit = read_counted_string_from_buffer(fp)
        readback_name = read_counted_string_from_buffer(fp)
        readback_description = read_counted_string_from_buffer(fp)
        readback_unit = read_counted_string_from_buffer(fp)

        return cls(
            number,
            name,
            description,
            step_mode,
            unit,
            readback_name,
            readback_description,
            readback_unit,
        )

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'number': self.number,
            'name': self.name,
            'description': self.description,
            'step_mode': self.step_mode,
            'unit': self.unit,
            'readback_name': self.readback_name,
            'readback_description': self.readback_description,
            'readback_unit': self.readback_unit,
        }


@dataclass(frozen=True)
class MDAScanDetectorInfo:
    number: int
    name: str
    description: str
    unit: str

    @classmethod
    def read(cls, fp: typing.BinaryIO) -> MDAScanDetectorInfo:
        number = read_int_from_buffer(fp)
        name = read_counted_string_from_buffer(fp)
        description = read_counted_string_from_buffer(fp)
        unit = read_counted_string_from_buffer(fp)
        return cls(number, name, description, unit)

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'number': self.number,
            'name': self.name,
            'description': self.description,
            'unit': self.unit,
        }


@dataclass(frozen=True)
class MDAScanTriggerInfo:
    number: int
    name: str
    command: float

    @classmethod
    def read(cls, fp: typing.BinaryIO) -> MDAScanTriggerInfo:
        number = read_int_from_buffer(fp)
        name = read_counted_string_from_buffer(fp)
        command = read_float_from_buffer(fp)
        return cls(number, name, command)

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'number': self.number,
            'name': self.name,
            'command': self.command,
        }


@dataclass(frozen=True)
class MDAScanInfo:
    scan_name: str
    time_stamp: str
    positioner: list[MDAScanPositionerInfo]
    detector: list[MDAScanDetectorInfo]
    trigger: list[MDAScanTriggerInfo]

    @classmethod
    def read(cls, fp: typing.BinaryIO) -> MDAScanInfo:
        scan_name = read_counted_string_from_buffer(fp)
        time_stamp = read_counted_string_from_buffer(fp)

        unpacker = xdrlib.Unpacker(fp.read(12))
        np = unpacker.unpack_int()
        nd = unpacker.unpack_int()
        nt = unpacker.unpack_int()

        positioner = [MDAScanPositionerInfo.read(fp) for p in range(np)]
        detector = [MDAScanDetectorInfo.read(fp) for d in range(nd)]
        trigger = [MDAScanTriggerInfo.read(fp) for t in range(nt)]

        return cls(scan_name, time_stamp, positioner, detector, trigger)

    @property
    def num_positioners(self) -> int:
        return len(self.positioner)

    @property
    def num_detectors(self) -> int:
        return len(self.detector)

    @property
    def num_triggers(self) -> int:
        return len(self.trigger)

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'scan_name': self.scan_name,
            'time_stamp': self.time_stamp,
            'positioner': [pos.to_mapping() for pos in self.positioner],
            'detector': [det.to_mapping() for det in self.detector],
            'trigger': [tri.to_mapping() for tri in self.trigger],
        }


@dataclass(frozen=True)
class MDAScanData:
    readback_array: RealArrayType  # double, shape: np x npts
    detector_array: RealArrayType  # float, shape: nd x npts

    @classmethod
    def read(
        cls, fp: typing.BinaryIO, scan_header: MDAScanHeader, scan_info: MDAScanInfo
    ) -> MDAScanData:
        npts = scan_header.num_requested_points
        np = scan_info.num_positioners
        nd = scan_info.num_detectors

        unpacker = xdrlib.Unpacker(fp.read(8 * np * npts))
        readback_lol = [unpacker.unpack_farray(npts, unpacker.unpack_double) for p in range(np)]
        readback_array = numpy.array(readback_lol)

        unpacker.reset(fp.read(4 * nd * npts))
        detector_lol = [unpacker.unpack_farray(npts, unpacker.unpack_float) for d in range(nd)]
        detector_array = numpy.array(detector_lol)

        return cls(readback_array, detector_array)

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'readback_array': f'{self.readback_array.dtype}{self.readback_array.shape}',
            'detector_array': f'{self.detector_array.dtype}{self.detector_array.shape}',
        }


@dataclass(frozen=True)
class MDAScan:
    header: MDAScanHeader
    info: MDAScanInfo
    data: MDAScanData
    lower_scans: list[MDAScan]

    @classmethod
    def read(cls, fp: typing.BinaryIO) -> MDAScan:
        header = MDAScanHeader.read(fp)
        info = MDAScanInfo.read(fp)
        data = MDAScanData.read(fp, header, info)
        lower_scans: list[MDAScan] = list()

        for offset in header.lower_scan_offsets:
            fp.seek(offset)
            scan = MDAScan.read(fp)
            lower_scans.append(scan)

        return cls(header, info, data, lower_scans)

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'header': self.header.to_mapping(),
            'info': self.info.to_mapping(),
            'data': self.data.to_mapping(),
            'lower_scans': [scan.to_mapping() for scan in self.lower_scans],
        }

    def __str__(self) -> str:
        return yaml.safe_dump(self.to_mapping(), sort_keys=False)


@dataclass(frozen=True)
class MDAProcessVariable(Generic[T]):
    name: str
    description: str
    epics_type: EpicsType
    unit: str
    value: T

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'name': self.name,
            'description': self.description,
            'epicsType': self.epics_type.name,
            'unit': self.unit,
            'value': self.value,
        }


@dataclass(frozen=True)
class MDAFile:
    header: MDAHeader
    scan: MDAScan
    extra_pvs: list[MDAProcessVariable[Any]]

    @staticmethod
    def _read_pv(unpacker: xdrlib.Unpacker) -> MDAProcessVariable[typing.Any]:
        pv_name = read_counted_string(unpacker)
        pv_desc = read_counted_string(unpacker)
        pv_type = EpicsType(unpacker.unpack_int())

        if pv_type == EpicsType.DBR_STRING:
            value_str = read_counted_string(unpacker)
            return MDAProcessVariable[str](pv_name, pv_desc, pv_type, str(), value_str)

        count = unpacker.unpack_int()
        pv_unit = read_counted_string(unpacker)

        if pv_type == EpicsType.DBR_CTRL_CHAR:
            value_char = unpacker.unpack_fstring(count).decode()
            value_char = value_char.split('\x00', 1)[0]  # treat as null-terminated string
            return MDAProcessVariable[str](pv_name, pv_desc, pv_type, pv_unit, value_char)
        elif pv_type == EpicsType.DBR_CTRL_SHORT:
            value_short = unpacker.unpack_farray(count, unpacker.unpack_int)
            return MDAProcessVariable[list[int]](pv_name, pv_desc, pv_type, pv_unit, value_short)
        elif pv_type == EpicsType.DBR_CTRL_LONG:
            value_long = unpacker.unpack_farray(count, unpacker.unpack_int)
            return MDAProcessVariable[list[int]](pv_name, pv_desc, pv_type, pv_unit, value_long)
        elif pv_type == EpicsType.DBR_CTRL_FLOAT:
            value_float = unpacker.unpack_farray(count, unpacker.unpack_float)
            return MDAProcessVariable[list[float]](pv_name, pv_desc, pv_type, pv_unit, value_float)
        elif pv_type == EpicsType.DBR_CTRL_DOUBLE:
            value_double = unpacker.unpack_farray(count, unpacker.unpack_double)
            return MDAProcessVariable[list[float]](pv_name, pv_desc, pv_type, pv_unit, value_double)

        return MDAProcessVariable[str](pv_name, pv_desc, pv_type, pv_unit, str())

    @classmethod
    def read(cls, file_path: Path) -> MDAFile:
        extra_pvs: list[MDAProcessVariable[Any]] = list()

        with file_path.open(mode='rb') as fp:
            header = MDAHeader.read(fp)
            scan = MDAScan.read(fp)

            if header.has_extra_pvs:
                fp.seek(header.extra_pvs_offset)
                unpacker = xdrlib.Unpacker(fp.read())
                number_pvs = unpacker.unpack_int()

                for pvidx in range(number_pvs):
                    pv = cls._read_pv(unpacker)
                    extra_pvs.append(pv)

        return cls(header, scan, extra_pvs)

    def to_mapping(self) -> Mapping[str, Any]:
        return {
            'header': self.header.to_mapping(),
            'scan': self.scan.to_mapping(),
            'extra_pvs': [pv.to_mapping() for pv in self.extra_pvs],
        }

    def __str__(self) -> str:
        return yaml.safe_dump(self.to_mapping(), sort_keys=False)


class MDAPositionFileReader(ProbePositionFileReader):
    def __init__(self, scale_to_meters: float) -> None:
        self._scale_to_meters = scale_to_meters

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        mda_file = MDAFile.read(file_path)

        yscan = mda_file.scan
        yarray = yscan.data.readback_array[0, :]

        for y, xscan in zip(yarray, yscan.lower_scans):
            xarray = xscan.data.readback_array[0, :]

            for x in xarray:
                point = ProbePosition(
                    index=len(point_list),
                    coordinate_x_m=float(x) * self._scale_to_meters,
                    coordinate_y_m=float(y) * self._scale_to_meters,
                )
                point_list.append(point)

        return ProbePositionSequence(point_list)


class MDAFlatScanPositionFileReader(ProbePositionFileReader):
    def __init__(self, scale_to_meters: float) -> None:
        self._scale_to_meters = scale_to_meters

    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list = list()

        mda_file = MDAFile.read(file_path)
        xarray = mda_file.scan.data.readback_array[0, :]
        yarray = mda_file.scan.data.readback_array[1, :]

        for idx, (x, y) in enumerate(zip(xarray, yarray)):
            point = ProbePosition(
                index=idx,
                coordinate_x_m=float(x) * self._scale_to_meters,
                coordinate_y_m=float(y) * self._scale_to_meters,
            )
            point_list.append(point)

        return ProbePositionSequence(point_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_position_file_readers.register_plugin(
        MDAPositionFileReader(scale_to_meters=1.0e-6),
        simple_name='MDA',
        display_name='EPICS MDA Files (*.mda)',
    )
    registry.probe_position_file_readers.register_plugin(
        MDAPositionFileReader(scale_to_meters=1.0e-3),
        simple_name='APS_2IDD',
        display_name='APS 2-ID-D Microprobe Files (*.mda)',
    )
    registry.probe_position_file_readers.register_plugin(
        MDAPositionFileReader(scale_to_meters=1.0e-3),
        simple_name='APS_2IDE',
        display_name='APS 2-ID-E Microprobe Files (*.mda)',
    )
    registry.probe_position_file_readers.register_plugin(
        MDAPositionFileReader(scale_to_meters=1.0e-6),
        simple_name='APS_BNP',
        display_name='APS 2-ID-D Bionanoprobe Files (*.h5 *.hdf5)',
    )
    registry.probe_position_file_readers.register_plugin(
        MDAFlatScanPositionFileReader(scale_to_meters=1.0e-3),
        simple_name='APS_ISN_MDA',
        display_name='APS 19-ID In-Situ Nanoprobe Files (*.mda)',
    )
    registry.probe_position_file_readers.register_plugin(
        MDAFlatScanPositionFileReader(scale_to_meters=1.0e-6),
        simple_name='CNM_APS_HXN',
        display_name='CNM/APS Hard X-ray Nanoprobe Files (*.mda)',
    )


if __name__ == '__main__':
    file_path = Path(sys.argv[1])
    mda_file = MDAFile.read(file_path)
    print(mda_file)



from collections.abc import Sequence
import logging
import requests

from .chat import ChatHistory, ChatMessage, ChatRole, ChatTerminal
from .settings import ArgoSettings

logger = logging.getLogger(__name__)


class ArgoChatTerminal(ChatTerminal):
    def __init__(self, settings: ArgoSettings, history: ChatHistory) -> None:
        self._settings = settings
        self._history = history

    def send_message(self, content: str, stop: Sequence[str] = []) -> None:
        if not content:
            return

        messages = [
            ChatMessage(
                role=ChatRole.SYSTEM, content='You are a large language model with the name Argo.'
            )
        ]

        for line in content.splitlines():
            message = ChatMessage(role=ChatRole.USER, content=line)
            messages.append(message)
            self._history.add_message(message)

        logger.debug(f'{messages=}')

        url = self._settings.chat_endpoint_url.get_value()
        payload = {
            'user': self._settings.user.get_value(),
            'model': self._settings.chat_model.get_value(),
            'messages': [m.to_dict() for m in messages],
            'stop': stop,
            'temperature': self._settings.temperature.get_value(),
            'top_p': self._settings.top_p.get_value(),
            'max_tokens': self._settings.max_tokens.get_value(),
            'max_completion_tokens': self._settings.max_completion_tokens.get_value(),
        }
        headers = {'Content-Type': 'application/json'}
        response = requests.post(url, json=payload, headers=headers)

        logger.debug(f'{response=}')
        logger.debug(f'Status Code: {response.status_code}')
        response_json = response.json()
        logger.debug(f'JSON Response: {response_json}')
        response.raise_for_status()

        response_message = ChatMessage(role=ChatRole.AGENT, content=response_json['response'])
        self._history.add_message(response_message)

    def embed_texts(self, texts: Sequence[str]) -> Sequence[Sequence[float]]:
        """Generates embeddings for a list of strings."""
        url = self._settings.embeddings_endpoint_url.get_value()
        payload = {
            'user': self._settings.user.get_value(),
            'model': self._settings.embeddings_model.get_value(),
            'prompt': texts,
        }
        headers = {'Content-Type': 'application/json'}
        response = requests.post(url, json=payload, headers=headers)

        logger.debug(response)
        logger.debug(f'Status Code: {response.status_code}')
        response_json = response.json()
        logger.debug(f'JSON Response: {response_json}')
        response.raise_for_status()

        return response_json['embedding']



from __future__ import annotations
from collections.abc import Iterator

import numpy

from PyQt5.QtCore import Qt, QPointF, QRectF, QSize
from PyQt5.QtGui import QColor, QConicalGradient, QIcon, QLinearGradient, QPainter, QPen
from PyQt5.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGridLayout,
    QHBoxLayout,
    QPushButton,
    QSizePolicy,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from ptychodus.api.common import RealArrayType

from .visualization import VisualizationView
from .widgets import BottomTitledGroupBox, DecimalLineEdit, DecimalSlider


class ImageDisplayRangeDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.button_box = QDialogButtonBox()
        self.min_value_line_edit = DecimalLineEdit.create_instance()
        self.max_value_line_edit = DecimalLineEdit.create_instance()

        self.setWindowTitle('Set Display Range')
        self.button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        self.button_box.accepted.connect(self.accept)
        self.button_box.addButton(QDialogButtonBox.StandardButton.Cancel)
        self.button_box.rejected.connect(self.reject)

        layout = QFormLayout()
        layout.addRow('Minimum Displayed Value:', self.min_value_line_edit)
        layout.addRow('Maximum Displayed Value:', self.max_value_line_edit)
        layout.addRow(self.button_box)
        self.setLayout(layout)


class ImageToolsGroupBox(BottomTitledGroupBox):
    def __init__(self, parent: QWidget | None = None, *, add_fourier_tool: bool = False) -> None:
        super().__init__('Tools', parent)
        self.home_button = QToolButton()
        self.save_button = QToolButton()
        self.move_button = QToolButton()
        self.fourier_button = QToolButton()
        self.ruler_button = QToolButton()
        self.rectangle_button = QToolButton()
        self.line_cut_button = QToolButton()

        self.home_button.setIcon(QIcon(':/icons/home'))
        self.home_button.setIconSize(QSize(32, 32))
        self.home_button.setToolTip('Home')

        self.save_button.setIcon(QIcon(':/icons/save'))
        self.save_button.setIconSize(QSize(32, 32))
        self.save_button.setToolTip('Save Image')

        self.move_button.setIcon(QIcon(':/icons/move'))
        self.move_button.setIconSize(QSize(32, 32))
        self.move_button.setToolTip('Move')

        self.fourier_button.setIcon(QIcon(':/icons/fourier'))
        self.fourier_button.setIconSize(QSize(32, 32))
        self.fourier_button.setToolTip('Fourier Transform')

        self.ruler_button.setIcon(QIcon(':/icons/ruler'))
        self.ruler_button.setIconSize(QSize(32, 32))
        self.ruler_button.setToolTip('Ruler')

        self.rectangle_button.setIcon(QIcon(':/icons/rectangle'))
        self.rectangle_button.setIconSize(QSize(32, 32))
        self.rectangle_button.setToolTip('Rectangle')

        self.line_cut_button.setIcon(QIcon(':/icons/line-cut'))
        self.line_cut_button.setIconSize(QSize(32, 32))
        self.line_cut_button.setToolTip('Line-Cut Profile')

        layout = QGridLayout()
        layout.addWidget(self.home_button, 0, 0)
        layout.addWidget(self.save_button, 0, 1)
        layout.addWidget(self.move_button, 0, 2)

        if add_fourier_tool:
            layout.addWidget(self.fourier_button, 0, 3)

        layout.addWidget(self.ruler_button, 1, 0)
        layout.addWidget(self.rectangle_button, 1, 1)
        layout.addWidget(self.line_cut_button, 1, 2)
        self.setLayout(layout)

        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)


class ImageRendererGroupBox(BottomTitledGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Colorize', parent)
        self.renderer_combo_box = QComboBox()
        self.transformation_combo_box = QComboBox()
        self.variant_combo_box = QComboBox()

        self.renderer_combo_box.setToolTip('Array Component')
        self.transformation_combo_box.setToolTip('Transformation')
        self.variant_combo_box.setToolTip('Variant')

        layout = QVBoxLayout()
        layout.setContentsMargins(10, 10, 10, 35)
        layout.addWidget(self.renderer_combo_box)
        layout.addWidget(self.transformation_combo_box)
        layout.addWidget(self.variant_combo_box)
        self.setLayout(layout)

        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)


class ImageDataRangeGroupBox(BottomTitledGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Data Range', parent)
        self.min_display_value_slider = DecimalSlider.create_instance(Qt.Orientation.Horizontal)
        self.max_display_value_slider = DecimalSlider.create_instance(Qt.Orientation.Horizontal)
        self.auto_button = QPushButton('Auto')
        self.edit_button = QPushButton('Edit')
        self.color_legend_button = QPushButton('Color Legend')

        self.min_display_value_slider.setToolTip('Minimum Display Value')
        self.max_display_value_slider.setToolTip('Maximum Display Value')
        self.auto_button.setToolTip('Rescale to Data Range')
        self.edit_button.setToolTip('Rescale to Custom Range')
        self.color_legend_button.setToolTip('Toggle Color Legend Visibility')

        button_layout = QHBoxLayout()
        button_layout.setContentsMargins(0, 0, 0, 0)
        button_layout.addWidget(self.auto_button)
        button_layout.addWidget(self.edit_button)
        button_layout.addWidget(self.color_legend_button)

        layout = QFormLayout()
        layout.setContentsMargins(10, 10, 10, 35)
        layout.addRow('Min:', self.min_display_value_slider)
        layout.addRow('Max:', self.max_display_value_slider)
        layout.addRow(button_layout)
        self.setLayout(layout)

        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)


class ImageRibbon(QWidget):
    def __init__(self, parent: QWidget | None = None, *, add_fourier_tool: bool = False) -> None:
        super().__init__(parent)
        self.image_tools_group_box = ImageToolsGroupBox(add_fourier_tool=add_fourier_tool)
        self.colormap_group_box = ImageRendererGroupBox()
        self.data_range_group_box = ImageDataRangeGroupBox()

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.image_tools_group_box)
        layout.addWidget(self.colormap_group_box)
        layout.addWidget(self.data_range_group_box)
        self.setLayout(layout)

        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)


class ImageWidget(VisualizationView):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._color_legend_min_value = 0.0
        self._color_legend_max_value = 1.0
        self._color_legend_stop_points: list[tuple[float, QColor]] = [
            (0.0, QColor(Qt.GlobalColor.green)),
            (0.5, QColor(Qt.GlobalColor.yellow)),
            (1.0, QColor(Qt.GlobalColor.red)),
        ]
        self._color_legend_num_ticks = 5  # TODO
        self._is_color_legend_visible = False
        self._is_color_legend_cyclic = False

    def set_color_legend_colors(
        self, values: RealArrayType, rgba_array: RealArrayType, is_cyclic: bool
    ) -> None:
        color_legend_stop_points: list[tuple[float, QColor]] = list()
        self._color_legend_min_value = values.min()
        self._color_legend_max_value = values.max()

        value_range = self._color_legend_max_value - self._color_legend_min_value
        normalized_values = (
            (values - self._color_legend_min_value) / value_range
            if value_range > 0
            else numpy.full_like(values, 0.5)
        )

        for x, rgba in zip(normalized_values.clip(0, 1), rgba_array):
            color = QColor()
            color.setRgbF(rgba[0], rgba[1], rgba[2], rgba[3])
            color_legend_stop_points.append((x, color))

        self._color_legend_stop_points = color_legend_stop_points
        self._is_color_legend_cyclic = is_cyclic
        scene = self.scene()

        if scene is not None:
            scene.update()

    def set_color_legend_visible(self, visible: bool) -> None:
        self._is_color_legend_visible = visible
        scene = self.scene()

        if scene is not None:
            scene.update()

    @property
    def _color_legend_ticks(self) -> Iterator[float]:
        for tick in range(self._color_legend_num_ticks):
            a = tick / (self._color_legend_num_ticks - 1)
            yield (1.0 - a) * self._color_legend_min_value + a * self._color_legend_max_value

    def drawForeground(self, painter: QPainter | None, rect: QRectF) -> None:  # noqa: N802
        if not self._is_color_legend_visible:
            return

        fg_painter = QPainter(self.viewport())

        pen = QPen()
        pen.setWidth(3)
        fg_painter.setPen(pen)

        font_metrics = fg_painter.fontMetrics()
        dx = font_metrics.horizontalAdvance('m')
        dy = font_metrics.lineSpacing()

        viewport = self.viewport()

        if viewport is None:
            return

        widget_rect = viewport.rect()

        if self._is_color_legend_cyclic:
            legend_diameter = 6 * dx
            legend_margin = 2 * dx

            legend_rect = QRectF(0.0, 0.0, legend_diameter, legend_diameter)
            legend_rect.moveRight(widget_rect.right() - legend_margin)
            legend_rect.moveBottom(widget_rect.height() - legend_margin)

            cgradient = QConicalGradient(legend_rect.center(), 90.0)
            cgradient.setStops(self._color_legend_stop_points)
            fg_painter.setBrush(cgradient)
            fg_painter.drawEllipse(legend_rect)
        else:
            tick_labels = [f'{tick:5g}' for tick in self._color_legend_ticks]
            tick_label_width = max(font_metrics.width(label) for label in tick_labels)

            legend_width = 2 * dx
            legend_height = (2 * len(tick_labels) - 1) * dy
            legend_margin = tick_label_width + 2 * dx

            legend_rect = QRectF(0.0, 0.0, legend_width, legend_height)
            legend_rect.moveRight(widget_rect.right() - legend_margin)
            legend_rect.moveTop((widget_rect.height() - legend_height) // 2)

            lgradient = QLinearGradient(legend_rect.bottomLeft(), legend_rect.topLeft())
            lgradient.setStops(self._color_legend_stop_points)
            fg_painter.setBrush(lgradient)
            fg_painter.drawRect(legend_rect)

            tick_x0 = legend_rect.right() + dx
            tick_y0 = legend_rect.bottom() + font_metrics.strikeOutPos()

            for tick_index, tick_label in enumerate(tick_labels):
                tick_dy = (tick_index * legend_rect.height()) // (len(tick_labels) - 1)
                viewport_point = QPointF(tick_x0, tick_y0 - tick_dy)
                fg_painter.drawText(viewport_point, tick_label)


class ImageView(QWidget):
    def __init__(self, parent: QWidget | None = None, *, add_fourier_tool: bool = False) -> None:
        super().__init__(parent)
        self.image_ribbon = ImageRibbon(add_fourier_tool=add_fourier_tool)
        self.image_widget = ImageWidget()

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setMenuBar(self.image_ribbon)
        layout.addWidget(self.image_widget)
        self.setLayout(layout)

        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)



from decimal import Decimal
import logging

from PyQt5.QtWidgets import QButtonGroup

from ptychodus.api.common import RealArrayType
from ptychodus.api.observer import Observable, Observer

from ...model.analysis import IlluminationMapper
from ...model.visualization import VisualizationEngine
from ...view.probe import IlluminationDialog, IlluminationParametersView, IlluminationQuantityView
from ...view.widgets import ExceptionDialog
from ..data import FileDialogFactory
from ..visualization import (
    VisualizationParametersController,
    VisualizationWidgetController,
)

logger = logging.getLogger(__name__)


class IlluminationParametersController(Observer):
    def __init__(self, mapper: IlluminationMapper, view: IlluminationParametersView) -> None:
        super().__init__()
        self._mapper = mapper
        self._view = view

        self._view.photon_flux_line_edit.setEnabled(False)
        self._view.exposure_time_line_edit.setEnabled(False)
        self._view.mass_attenuation_line_edit.setEnabled(False)

        self._mapper.add_observer(self)

    def _update_parameters(self) -> None:
        try:
            illumination_map = self._mapper.get_illumination_map()
        except ValueError:
            pass
        else:
            self._view.photon_flux_line_edit.set_value(
                Decimal(repr(illumination_map.photon_flux_Hz))
            )
            self._view.exposure_time_line_edit.set_value(
                Decimal(repr(illumination_map.exposure_time_s))
            )
            self._view.mass_attenuation_line_edit.set_value(
                Decimal(repr(illumination_map.mass_attenuation_m2_kg))
            )
            return

        nan = Decimal('NaN')
        self._view.exposure_time_line_edit.set_value(nan)
        self._view.mass_attenuation_line_edit.set_value(nan)

    def _update(self, observable: Observable) -> None:
        if observable is self._mapper:
            self._update_parameters()


class IlluminationQuantityController(Observer):
    def __init__(
        self,
        mapper: IlluminationMapper,
        view: IlluminationQuantityView,
        widget_controller: VisualizationWidgetController,
    ) -> None:
        super().__init__()
        self._mapper = mapper
        self._view = view
        self._widget_controller = widget_controller

        self._button_group = QButtonGroup()
        self._button_group.addButton(view.photon_number_button)
        self._button_group.addButton(view.photon_fluence_button)
        self._button_group.addButton(view.photon_fluence_rate_button)
        self._button_group.addButton(view.energy_fluence_button)
        self._button_group.addButton(view.energy_fluence_rate_button)
        self._button_group.addButton(view.dose_button)
        self._button_group.addButton(view.dose_rate_button)
        self._button_group.setExclusive(True)
        view.photon_number_button.setChecked(True)

        self._button_group.buttonClicked.connect(self._update_quantity)
        self._mapper.add_observer(self)

    def _update_quantity(self) -> None:
        try:
            illumination_map = self._mapper.get_illumination_map()
        except ValueError:
            self._widget_controller.clear_array()
            return

        quantity: RealArrayType | None = None

        match self._button_group.checkedButton():
            case self._view.photon_number_button:
                quantity = illumination_map.photon_number
            case self._view.photon_fluence_button:
                quantity = illumination_map.photon_fluence_1_m2
            case self._view.photon_fluence_rate_button:
                quantity = illumination_map.photon_fluence_rate_Hz_m2
            case self._view.energy_fluence_button:
                quantity = illumination_map.energy_fluence_J_m2
            case self._view.energy_fluence_rate_button:
                quantity = illumination_map.energy_fluence_rate_W_m2
            case self._view.dose_button:
                quantity = illumination_map.dose_Gy
            case self._view.dose_rate_button:
                quantity = illumination_map.dose_rate_Gy_s

        if quantity is None:
            self._widget_controller.clear_array()
        else:
            self._widget_controller.set_array(quantity, illumination_map.pixel_geometry)

    def _update(self, observable: Observable) -> None:
        if observable is self._mapper:
            self._update_quantity()


class IlluminationViewController:
    def __init__(
        self,
        mapper: IlluminationMapper,
        engine: VisualizationEngine,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._mapper = mapper
        self._file_dialog_factory = file_dialog_factory
        self._dialog = IlluminationDialog()

        self._parameters_controller = IlluminationParametersController(
            mapper, self._dialog.parameters_view
        )
        self._visualization_widget_controller = VisualizationWidgetController(
            engine,
            self._dialog.visualization_widget,
            self._dialog.status_bar,
            file_dialog_factory,
        )
        self._quantity_controller = IlluminationQuantityController(
            mapper, self._dialog.quantity_view, self._visualization_widget_controller
        )
        self._visualization_parameters_controller = VisualizationParametersController(
            engine, self._dialog.visualization_parameters_view
        )
        self._dialog.save_button.clicked.connect(self._save_data)

    def map(self, product_index: int) -> None:
        self._mapper.set_product(product_index)

        try:
            product_name = self._mapper.get_product_name()
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Illumination Mapper', err)
        else:
            self._dialog.setWindowTitle(f'Illumination Map: {product_name}')
            self._dialog.open()

        try:
            self._mapper.map()
        except Exception as err:
            logger.exception(err)
            ExceptionDialog.show_exception('Illumination Mapper', err)

    def _save_data(self) -> None:
        title = 'Save Illumination Map'
        file_path, _ = self._file_dialog_factory.get_save_file_path(
            self._dialog,
            title,
            name_filters=self._mapper.get_save_file_filters(),
            selected_name_filter=self._mapper.get_save_file_filter(),
        )

        if file_path:
            try:
                self._mapper.save_data(file_path)
            except Exception as err:
                logger.exception(err)
                ExceptionDialog.show_exception(title, err)



"""Visualization data containers for plots, line cuts, and colorized image products."""

from __future__ import annotations
from collections.abc import Iterator, Sequence
from dataclasses import dataclass
from typing import Final

from scipy.stats import gaussian_kde
import numpy

from .common import NumberArrayType, RealArrayType
from .geometry import Box2D, Interval, Line2D, PixelGeometry


@dataclass(frozen=True)
class PlotSeries:
    """A named sequence of float values representing one data series in a plot."""

    label: str
    values: Sequence[float]

    def copy(self) -> PlotSeries:
        return PlotSeries(str(self.label), list(self.values))


@dataclass(frozen=True)
class PlotAxis:
    """A labeled axis containing one or more PlotSeries."""

    label: str
    series: Sequence[PlotSeries]

    @classmethod
    def create_null(cls) -> PlotAxis:
        return cls('', [])

    def copy(self) -> PlotAxis:
        return PlotAxis(str(self.label), [series.copy() for series in self.series])


@dataclass(frozen=True)
class Plot2D:
    """A 2D plot with labeled X and Y axes, each containing one or more data series."""

    axis_x: PlotAxis
    axis_y: PlotAxis

    @classmethod
    def create_null(cls) -> Plot2D:
        return cls(PlotAxis.create_null(), PlotAxis.create_null())

    def copy(self) -> Plot2D:
        return Plot2D(self.axis_x.copy(), self.axis_y.copy())


@dataclass(frozen=True)
class LineCut:
    """1D profile sampled along a line: distances in meters and corresponding pixel values."""

    distance_m: Sequence[float]
    value: Sequence[float | complex]


@dataclass(frozen=True)
class KernelDensityEstimate:
    """KDE of pixel values within a region, including the value range and fitted kde object."""

    value_lower: float
    value_upper: float
    kde: gaussian_kde


class VisualizationProduct:
    """Colorized image with associated scalar values and pixel geometry for display and analysis."""

    EPS: Final[float] = 1.0e-6

    def __init__(
        self,
        value_label: str,
        values: NumberArrayType,
        rgba: RealArrayType,
        pixel_geometry: PixelGeometry,
    ) -> None:
        if values.ndim != 2:
            raise ValueError(f'Values must be a 2-dimensional ndarray (actual={values.ndim}).')

        if rgba.ndim != 3:
            raise ValueError(f'RGBA must be a 3-dimensional ndarray (actual={rgba.ndim}).')

        if rgba.shape[2] != 4:
            raise ValueError(f'RGBA final dimension must have length=4 (actual={rgba.shape[2]}).')

        if values.shape[0] != rgba.shape[0] or values.shape[1] != rgba.shape[1]:
            raise ValueError(f'Shape mismatch (values={values.shape} and rgba={rgba.shape}).')

        self._value_label = value_label
        self._values = values
        self._rgba = rgba
        self._pixel_width_m = pixel_geometry.width_m
        self._pixel_height_m = pixel_geometry.height_m

    def get_value_label(self) -> str:
        return self._value_label

    def get_values(self) -> NumberArrayType:
        return self._values

    def get_image_rgba(self) -> RealArrayType:
        return self._rgba

    def get_pixel_geometry(self) -> PixelGeometry:
        return PixelGeometry(
            width_m=self._pixel_width_m,
            height_m=self._pixel_height_m,
        )

    @staticmethod
    def _intersect_bounding_box(begin: float, end: float, n: int) -> Interval[float]:
        length = end - begin

        if abs(length) < VisualizationProduct.EPS:
            return Interval[float](-numpy.inf, numpy.inf)
        else:
            return Interval[float].create_proper(
                (0 - begin) / length,
                (n - begin) / length,
            )

    @staticmethod
    def _intersect_grid_lines(
        begin: float, end: float, alpha_limits: Interval[float]
    ) -> Iterator[float]:
        ibegin = int(begin)
        iend = int(end)

        if iend < ibegin:
            ibegin, iend = iend, ibegin

        length = end - begin

        if abs(length) > VisualizationProduct.EPS:
            for idx in range(ibegin, iend + 1):
                alpha = (idx - begin) / length

                if alpha in alpha_limits:
                    yield alpha

    def _clip_to_bounding_box(self, line: Line2D) -> Interval[float]:
        alpha_x = self._intersect_bounding_box(line.begin.x, line.end.x, self._values.shape[-1])
        alpha_y = self._intersect_bounding_box(line.begin.y, line.end.y, self._values.shape[-2])

        return Interval[float].create_proper(
            max(0.0, max(alpha_x.lower, alpha_y.lower)),
            min(1.0, min(alpha_x.upper, alpha_y.upper)),
        )

    def _intersect_grid(self, line: Line2D) -> Sequence[float]:
        alpha_limits = self._clip_to_bounding_box(line)
        x_intersections = [
            x for x in self._intersect_grid_lines(line.begin.x, line.end.x, alpha_limits)
        ]
        y_intersections = [
            x for x in self._intersect_grid_lines(line.begin.y, line.end.y, alpha_limits)
        ]

        alpha = {alpha_limits.lower, alpha_limits.upper}
        alpha = alpha.union(x_intersections)
        alpha = alpha.union(y_intersections)
        return sorted(alpha)

    def get_info_text(self, x: float, y: float) -> str:
        ix = 0 if x < 0.0 else int(x)
        ix = min(ix, self._values.shape[-1])
        iy = 0 if y < 0.0 else int(y)
        iy = min(iy, self._values.shape[-2])
        value = self._values[iy, ix]

        if numpy.iscomplex(value):
            amplitude = numpy.absolute(value)
            phase = numpy.angle(value)
            return f'{x=:.1f} {y=:.1f} {amplitude=:6g} {phase=:6g}'

        return f'{x=:.1f} {y=:.1f} {value=:6g}'

    def get_line_cut(self, line: Line2D) -> LineCut:
        intersections = self._intersect_grid(line)

        dx = (line.end.x - line.begin.x) * self._pixel_width_m
        dy = (line.end.y - line.begin.y) * self._pixel_height_m
        line_length = numpy.hypot(dx, dy)

        distances: list[float] = list()
        values: list[float] = list()

        for alpha_l, alpha_r in zip(intersections[:-1], intersections[1:]):
            alpha = (alpha_l + alpha_r) / 2.0
            point = line.lerp(alpha)
            value = self._values[int(point.y), int(point.x)]

            distances.append(alpha * line_length)
            values.append(value)

        return LineCut(distances, values)

    def estimate_kernel_density(self, box: Box2D) -> KernelDensityEstimate:
        x_range = Interval[int](0, self._values.shape[-1])
        x_begin = x_range.clamp(int(box.x_begin))
        x_end = x_range.clamp(int(box.x_end) + 1)

        y_range = Interval[int](0, self._values.shape[-2])
        y_begin = y_range.clamp(int(box.y_begin))
        y_end = y_range.clamp(int(box.y_end) + 1)

        values = self._values[..., y_begin:y_end, x_begin:x_end]
        values = values.reshape(values.shape[-3], -1) if values.ndim > 2 else values.reshape(-1)

        if numpy.iscomplexobj(values):
            # TODO improve KDE for complex values
            values = numpy.absolute(values)

        return KernelDensityEstimate(values.min(), values.max(), gaussian_kde(values))
